/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/
#include "PLA.h"
#include "PLA_Blas1.h"

static PLA_Obj x_in = NULL, y_in = NULL;

int PLA_Swap_enter ( PLA_Obj x, PLA_Obj y )
{
   int instrument, check_parameter, check_seq, myrow, mycol;
   char current_routine[20] = "PLA_";
   int objtype;
   MPI_Datatype datatype;
   PLA_Template template;
   int  error = 0, original_error = 0;

   PLA_Stack_push_current( current_routine );

   PLA_Get_check_parameters ( &check_parameter );

   if ( PLA_CHECK_PARAMETER == check_parameter )
   {

    /* check objects for validity */
      PLA_Obj_object_valid( x,      "x", error, original_error );
      PLA_Obj_object_valid( y,      "y", error, original_error );

    /* check datatypes */
      PLA_Obj_datatype_valid( x, "x", error, original_error );
      PLA_Obj_datatype( x, &datatype );
      PLA_Obj_datatype_equal( y,     datatype, "y", error, original_error );

    /* check if templates are all equal (temporary restriction) */
      PLA_Obj_template( x, &template );
      PLA_Obj_template_equal( y,     template, "y", error, original_error );

      PLA_Obj_global_length_length_equal( x, y, "x and y", error, original_error );
      PLA_Obj_global_width_width_equal  ( x, y, "x and y", error, original_error );
   }

   PLA_Get_check_sequential ( &check_seq );
   if ( PLA_CHECK_SEQUENTIAL == check_seq )
   {
       PLA_Mscalar_create_conf_to( x,     0, 0, &x_in );
       PLA_Mscalar_create_conf_to( y,     0, 0, &y_in );

       PLA_Copy_temp( x,     x_in );
       PLA_Copy_temp( y,     y_in );

       PLA_Temp_comm_col_rank( template, &myrow );
       PLA_Temp_comm_row_rank( template, &mycol );

       if ( 0 == myrow && 0 == mycol )
          PLA_Local_inv_scal( x_in, y_in );
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
}


int PLA_Swap_exit ( PLA_Obj x, PLA_Obj y )
{
  char current_routine[20];
  int  check_seq, instrument, myrow, mycol;
  PLA_Obj y_out = NULL, x_out = NULL;
  PLA_Template template;

  PLA_Get_check_sequential ( &check_seq );
/* MAY (JAG) want to test the objects for validity here as well      */
/* in case we have a wayward pointer or some other memory corruption */
  if ( PLA_CHECK_SEQUENTIAL == check_seq )
  {
     PLA_Mscalar_create_conf_to( x,     0, 0, &x_out );
     PLA_Copy_temp( x, x_out );
     PLA_Mscalar_create_conf_to( y,     0, 0, &y_out );
     PLA_Copy_temp( y, y_out );
     PLA_Obj_template(x, &template);
     PLA_Temp_comm_col_rank( template, &myrow );
     PLA_Temp_comm_row_rank( template, &mycol );
     if ( 0 == myrow && 0 == mycol )
     {
/* Commented out because I know we have the code to do this somewhere */
/* and I would like to avoid re-writing it.  It's not a big deal to   */
/* rewrite if we restrict what can be passed to the function though.  */
/*        diff = PLA_max_abs_diff( x_in, y_out );
          diff = PLA_max_abs_diff( y_in, x_out );                     */
      }
   }

   PLA_Get_instrument ( &instrument );
   if ( PLA_INSTRUMENT == instrument )
   {
     /* insert instrumentation calls */
   }
   PLA_Stack_pop_current( current_routine );
   /* Free all of the "in" objects */
   PLA_Obj_free( &x_in  );
   PLA_Obj_free( &y_in  );
   PLA_Obj_free( &x_out );
   PLA_Obj_free( &y_out );
   return(PLA_SUCCESS);
}

/********************************************************************/
/* PLA_Swap : Swaps two PLA_Obj via PLA_Copy                        */
/********************************************************************/
int PLA_Swap    (PLA_Obj x,
                 PLA_Obj y)
/* I realize that this won't quite work -- but exactly what needs to be */
/* added depends on the precise state of the PLA_Copy routine.          */

{

   PLA_Obj temp = NULL;


   PLA_Swap_enter( x, y );

   PLA_Obj_clone(x, &temp);
   PLA_Copy( x, temp );
   PLA_Copy( y, x );
   PLA_Copy( temp, y );


   PLA_Obj_free( &temp );
   PLA_Swap_exit( x, y );
   return 0;
}
