#include "PLA.h"

static PLA_Obj alpha_in = NULL;
static PLA_Obj x_in     = NULL;
static PLA_Obj a_in     = NULL;

int PLA_Local_syr_enter(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	int check_parameter, instrument;
	char current_routine[20] = "PLA_Local_syr";
	int objtype;
	MPI_Datatype datatype;
	PLA_Template template;
	int  error = 0, original_error = 0;

	PLA_Stack_push_current( current_routine );

	PLA_Get_check_parameters ( &check_parameter );

	if ( PLA_CHECK_PARAMETER == check_parameter )
	{
	/* check uplo */
		if ( PLA_UPPER_TRIANGULAR != uplo &&
			PLA_LOWER_TRIANGULAR != uplo)
			PLA_Log_abort("PLA_Local_syr_enter", "PLA_Local_syr : uplo invalid", PLA_ERROR_CODE, original_error);

	/* check objects for validity */
		PLA_Obj_object_valid( alpha, "alpha", error, original_error);
		PLA_Obj_object_valid( x, "x", error, original_error);
		PLA_Obj_object_valid( a, "a", error, original_error);

	 /* check datatypes */
		PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
		PLA_Obj_datatype( alpha, &datatype );
		PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );
		PLA_Obj_datatype_equal( a, datatype, "a", error, original_error );

	 /* check if templates are all equal (temporary restriction) */
		PLA_Obj_template( alpha, &template );
		PLA_Obj_template_equal( x, template, "x", error, original_error );
		PLA_Obj_template_equal( a, template, "a", error, original_error );

	 /* check dimensions */
	  PLA_Obj_local_length_equal(alpha, 1, "alpha", error, original_error);
	  PLA_Obj_local_width_equal (alpha, 1, "alpha", error, original_error);
	  PLA_Obj_local_width_equal (x,     1, "x",     error, original_error);
		PLA_Obj_local_length_length_equal( x, a, "x and a", error, original_error );
		PLA_Obj_local_length_width_equal ( x, a, "x and a", error, original_error );
	}
	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	return(original_error);
}

int PLA_Local_syr_exit(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	int instrument;
	char current_routine[20];

	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	PLA_Stack_pop_current( current_routine );
	/* Free all of the "in" objects */
	return(PLA_SUCCESS);
}



int PLA_Local_syr(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	char uplo_char;
	int  error_level = 0;
	int  incx;
	int  lda, local_width_a, local_length_a, n;
	int  local_width_x, local_length_x;
	int  local_width_alpha, local_length_alpha;
	MPI_Datatype datatype_x, datatype_a, datatype_alpha;
	void  *a_buffer, *x_buffer, *alpha_buffer;

	PLA_Local_syr_enter(uplo, alpha, x, a);

	if(PLA_UPPER_TRIANGULAR == uplo) uplo_char = 'U';
	else uplo_char = 'L';

	PLA_Obj_local_length(a, &local_length_a);
	PLA_Obj_local_width(a, &local_width_a);
	PLA_Obj_local_ldim(a, &lda);
	PLA_Obj_local_buffer(a, &a_buffer);
	PLA_Obj_datatype(a, &datatype_a);

	PLA_Obj_local_length(x, &local_length_x);
	PLA_Obj_local_width(x, &local_width_x);
	PLA_Obj_local_stride(x, &incx);
	PLA_Obj_local_buffer(x, &x_buffer);
	PLA_Obj_datatype(x, &datatype_x);

	PLA_Obj_local_length(alpha, &local_length_alpha);
	PLA_Obj_local_width(alpha, &local_width_alpha);
	PLA_Obj_local_buffer(alpha, &alpha_buffer);
	PLA_Obj_datatype(alpha, &datatype_alpha);

	if(local_length_a * local_width_a > 0)
	{
		if(MPI_DOUBLE == datatype_a)
			PLA_dsyr(&uplo_char, &local_length_a, (double*)alpha_buffer,
						  (double*)x_buffer, &incx,(double*)a_buffer, &lda);
		else if(MPI_FLOAT == datatype_a)
			PLA_ssyr(&uplo_char, &local_length_a, (float*)alpha_buffer,
						(float*)x_buffer, &incx,(float*)a_buffer, &lda);
		if(MPI_DOUBLE_COMPLEX == datatype_a)
			PLA_zher(&uplo_char, &local_length_a, (PLA_DOUBLE_COMPLEX*)alpha_buffer,
						  (PLA_DOUBLE_COMPLEX*)x_buffer, &incx,(PLA_DOUBLE_COMPLEX*)a_buffer, &lda);
		else if(MPI_COMPLEX == datatype_a)
			PLA_cher(&uplo_char, &local_length_a, (PLA_COMPLEX*)alpha_buffer,
						(PLA_COMPLEX*)x_buffer, &incx,(PLA_COMPLEX*)a_buffer, &lda);
	}

	PLA_Local_syr_exit( uplo, alpha, x, a);
	return(PLA_SUCCESS);
}


int PLA_Syr_enter(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	int check_parameter, check_seq, instrument;
	int  myrow, mycol;
	char current_routine[20] = "PLA_Syr";
	int objtype;
	MPI_Datatype datatype;
	PLA_Template template;
	int  error = 0, original_error = 0;

	PLA_Stack_push_current( current_routine );

	PLA_Get_check_parameters ( &check_parameter );

	if ( PLA_CHECK_PARAMETER == check_parameter )
	{
	/* check uplo */
		if ( PLA_UPPER_TRIANGULAR != uplo &&
			PLA_LOWER_TRIANGULAR != uplo)
			PLA_Log_abort("PLA_Syr_enter", "PLA_Syr : uplo invalid", PLA_ERROR_CODE, original_error);

	/* check objects for validity */
		PLA_Obj_object_valid( alpha, "alpha", error, original_error);
		PLA_Obj_object_valid( x, "x", error, original_error);
		PLA_Obj_object_valid( a, "a", error, original_error);

	 /* check datatypes */
		PLA_Obj_datatype_valid( alpha, "alpha", error, original_error );
		PLA_Obj_datatype( alpha, &datatype );
		PLA_Obj_datatype_equal( x, datatype, "x", error, original_error );
		PLA_Obj_datatype_equal( a, datatype, "a", error, original_error );
	 /* check object types */
		PLA_Obj_objtype_equal( alpha, PLA_MSCALAR, "alpha", error, original_error );
	  PLA_Obj_objtype_equal( x,     PLA_ANY_VECTOR, "x",  error, original_error );
		PLA_Obj_objtype_equal( a,     PLA_MATRIX,  "a", error, original_error );

	 /* check if templates are all equal (temporary restriction) */
		PLA_Obj_template( alpha, &template );
		PLA_Obj_template_equal( x, template, "x", error, original_error );
		PLA_Obj_template_equal( a, template, "a", error, original_error );

	 /* check dimensions */
	  PLA_Obj_global_length_equal(alpha, 1, "alpha", error, original_error);
	  PLA_Obj_global_width_equal (alpha, 1, "alpha", error, original_error);
	  PLA_Obj_global_width_equal (x,     1, "x",     error, original_error);
		PLA_Obj_global_length_length_equal( x, a, "x and a", error, original_error );
		PLA_Obj_global_length_width_equal ( x, a, "x and a", error, original_error );
	}
	PLA_Get_check_sequential ( &check_seq );
	if ( PLA_CHECK_SEQUENTIAL == check_seq )
	{
		 PLA_Mscalar_create_conf_to( alpha, 0, 0, &alpha_in );
		 PLA_Mscalar_create_conf_to( x,     0, 0, &x_in );
		 PLA_Mscalar_create_conf_to( a,     0, 0, &a_in );

		 PLA_Copy_temp( alpha, alpha_in );
		 PLA_Copy_temp( x,     x_in );
		 PLA_Copy_temp( a,     a_in );

		 PLA_Temp_comm_col_rank( template, &myrow );
		 PLA_Temp_comm_row_rank( template, &mycol );

		 if ( 0 == myrow && 0 == mycol )
			 PLA_Local_syr( uplo, alpha_in, x_in, a_in );
	}
	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	return(original_error);
}

int PLA_Syr_exit(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	int myrow, mycol, instrument, check_seq;
	char current_routine[20];
	PLA_Obj a_out = NULL;
	PLA_Template template;

	PLA_Get_instrument ( &instrument );
	if ( PLA_INSTRUMENT == instrument )
	{
	  /* insert instrumentation calls */
	}
	PLA_Get_check_sequential ( &check_seq );

  if ( PLA_CHECK_SEQUENTIAL == check_seq )
  {
	  PLA_Mscalar_create_conf_to( a,     0, 0, &a_out );
	  PLA_Copy_temp(a, a_out);
	  PLA_Obj_template(a, &template);
	  PLA_Temp_comm_col_rank( template, &myrow );
	  PLA_Temp_comm_row_rank( template, &mycol );
	  if ( 0 == myrow && 0 == mycol )
	  {
/* Commented out because I know we have the code to do this somewhere */
/* and I would like to avoid re-writing it.  It's not a big deal to   */
/* rewrite if we restrict what can be passed to the function though.  */
/*        diff = PLA_max_abs_diff( a_in, a_out ); */
		}
	  PLA_Obj_free( &alpha_in);
	  PLA_Obj_free( &x_in);
	  PLA_Obj_free( &a_in);
	  PLA_Obj_free( &a_out);
	}

	PLA_Stack_pop_current( current_routine );
	return(PLA_SUCCESS);
}

int PLA_Syr_upper( PLA_Obj alpha,
						 PLA_Obj x_duplicated_col,
						 PLA_Obj x_duplicated_row,
						 PLA_Obj a)
{
	PLA_Obj x_row_left = NULL, x_row_right = NULL;
	PLA_Obj x_col_top = NULL, x_col_bottom = NULL;
	PLA_Obj a_left = NULL;
	PLA_Obj a_right_top = NULL, a_right_bottom = NULL;

	int right_size, right_owner;
	int bottom_size, bottom_owner;
	int size, my_proc_row, my_proc_col;
	PLA_Template template;

	PLA_Obj_template(a, &template);
	PLA_Temp_comm_col_rank( template, &my_proc_row );
	PLA_Temp_comm_row_rank( template, &my_proc_col );

	PLA_Obj_view_all( a, &a_left);
	PLA_Obj_view_all( x_duplicated_row, &x_row_left);
	PLA_Obj_view_all( x_duplicated_col, &x_col_top );
	while(TRUE)
	{
		PLA_Obj_split_size( a_left, PLA_SIDE_RIGHT, &right_size, &right_owner);
		PLA_Obj_split_size( a_left, PLA_SIDE_BOTTOM,&bottom_size,&bottom_owner );
		if( 0 == (size = min(right_size, bottom_size))) break;
		PLA_Obj_split_4(a_left,-size,-size, &a_left,  &a_right_top,
						    PLA_DUMMY, &a_right_bottom);
		PLA_Obj_vert_split_2(x_row_left,-size, &x_row_left, &x_row_right);
		PLA_Obj_horz_split_2(x_col_top, -size, &x_col_top,
															&x_col_bottom);
		if( my_proc_col == right_owner )
		{
			if( my_proc_row == bottom_owner )
			{
				PLA_Local_syr(PLA_UPPER_TRIANGULAR, alpha, x_col_bottom, a_right_bottom);
			}
			PLA_Local_ger(alpha, x_col_top, x_row_right, a_right_top);
		}
	}
	PLA_Obj_free( &x_row_left);
	PLA_Obj_free( &x_row_right);
	PLA_Obj_free( &x_col_top);
	PLA_Obj_free( &x_col_bottom);
	PLA_Obj_free( &a_left);
	PLA_Obj_free( &a_right_top);
	PLA_Obj_free( &a_right_bottom);
	return PLA_SUCCESS;
}


int PLA_Syr_lower( PLA_Obj alpha,
						 PLA_Obj x_duplicated_col,
						 PLA_Obj x_duplicated_row,
						 PLA_Obj a)
{
	PLA_Obj x_row_left = NULL, x_row_right = NULL;
	PLA_Obj x_col_top = NULL, x_col_bottom = NULL;
	PLA_Obj a_right = NULL;
	PLA_Obj a_left_top = NULL, a_left_bottom = NULL;

	int left_size, left_owner;
	int top_size, top_owner;
	int size, my_proc_row, my_proc_col;
	PLA_Template template;

	PLA_Obj_template(a, &template);
	PLA_Temp_comm_col_rank( template, &my_proc_row );
	PLA_Temp_comm_row_rank( template, &my_proc_col );

	PLA_Obj_view_all( a, &a_right);
	PLA_Obj_view_all( x_duplicated_row, &x_row_right);
	PLA_Obj_view_all( x_duplicated_col, &x_col_bottom );
	while(TRUE)
	{
		PLA_Obj_split_size( a_right, PLA_SIDE_LEFT, &left_size, &left_owner);
		PLA_Obj_split_size( a_right, PLA_SIDE_TOP,  &top_size,  &top_owner );
		if( 0 == (size = min(left_size, top_size))) break;
		PLA_Obj_split_4 ( a_right, size, size, & a_left_top,   PLA_DUMMY,
															&a_left_bottom, &a_right);
		PLA_Obj_vert_split_2(x_row_right,  size, &x_row_left, &x_row_right);
		PLA_Obj_horz_split_2(x_col_bottom, size, &x_col_top,
															  &x_col_bottom);
		if( my_proc_col == left_owner )
		{
			if( my_proc_row == top_owner )
			{
				PLA_Local_syr(PLA_LOWER_TRIANGULAR, alpha, x_col_top, a_left_top);
			}
			PLA_Local_ger(alpha, x_col_bottom, x_row_left, a_left_bottom);
		}
	}
	PLA_Obj_free( &x_row_left);
	PLA_Obj_free( &x_row_right);
	PLA_Obj_free( &x_col_top);
	PLA_Obj_free( &x_col_bottom);
	PLA_Obj_free( &a_right);
	PLA_Obj_free( &a_left_top);
	PLA_Obj_free( &a_left_bottom);
	return PLA_SUCCESS;
}

int PLA_Syr(int uplo, PLA_Obj alpha, PLA_Obj x, PLA_Obj a)
{
	int  error_level = 0;
	PLA_Obj x_duplicated_column = NULL, x_duplicated_row = NULL,
					 alpha_dup = NULL;
	int owner_row, owner_col;

	PLA_Syr_enter(uplo, alpha, x, a);

	PLA_Obj_owner_row(alpha, &owner_row);
	PLA_Obj_owner_col(alpha, &owner_col);
	if ( owner_row != PLA_ALL_ROWS || owner_col != PLA_ALL_COLS )
	{
		PLA_Mscalar_create_conf_to( alpha, PLA_ALL_ROWS, PLA_ALL_COLS, &alpha_dup);
		PLA_Copy( alpha, alpha_dup );
	}
	PLA_Pvector_create_conf_to(a, PLA_PROJ_ONTO_COL, PLA_ALL_COLS, &x_duplicated_column);
	PLA_Pvector_create_conf_to(a, PLA_PROJ_ONTO_ROW, PLA_ALL_ROWS, &x_duplicated_row);
	PLA_Copy(x, x_duplicated_column);
	PLA_Copy(x, x_duplicated_row);
	if(PLA_UPPER_TRIANGULAR == uplo)
	/* careful of the != == test here -- look in other code to see */
		PLA_Syr_upper( NULL == alpha_dup ? alpha : alpha_dup,
							x_duplicated_column,
							x_duplicated_row,
							a);
	else
		PLA_Syr_lower( NULL == alpha_dup ? alpha : alpha_dup,
							x_duplicated_column,
							x_duplicated_row,
							a);
	PLA_Obj_free( &alpha_dup);
	PLA_Obj_free( &x_duplicated_column);
	PLA_Obj_free( &x_duplicated_row);
	PLA_Syr_exit( uplo, alpha, x, a);
	return(PLA_SUCCESS);
}

