/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "PLA.h"
#include "PLA_Gemm.h"

/********************************************************/
/* PLA_GEMM()                                            */
/********************************************************/
int  PLA_Gemm_front (char *transa, char *transb, double *alpha,
					PLA_Obj a,
					PLA_Obj b,
					double *beta,
					PLA_Obj c )
{
	PLA_Obj Alpha_Mscalar = NULL, Beta_Mscalar = NULL;
	MPI_Datatype   Cs_datatype;
        PLA_Template   C_Template;
        int pla_transa, pla_transb;
        
        PLA_Set_check_parameters(PLA_NO_CHECK_PARAMETER);

        PLA_Obj_template(c, &C_Template);
	PLA_Obj_datatype(c, &Cs_datatype);
	PLA_Mscalar_create(Cs_datatype, PLA_ALL_ROWS, PLA_ALL_COLS,
                           1, 1, C_Template, &Alpha_Mscalar);
	PLA_Obj_set_contents(Alpha_Mscalar, alpha);
	PLA_Mscalar_create(Cs_datatype, PLA_ALL_ROWS, PLA_ALL_COLS,
                           1, 1, C_Template, &Beta_Mscalar);
	PLA_Obj_set_contents(Beta_Mscalar, beta);

        if((*transa == 'N') || (*transa == 'n'))
           pla_transa = PLA_NO_TRANSPOSE;
        else if((*transa == 'T') || (*transa == 't'))
           pla_transa = PLA_TRANSPOSE;
        else pla_transa = PLA_CONJ_TRANS;

        if((*transb == 'N') || (*transb == 'n'))
           pla_transb = PLA_NO_TRANSPOSE;
        else if((*transb == 'T') || (*transb == 't'))
           pla_transb = PLA_TRANSPOSE;
        else pla_transb = PLA_CONJ_TRANS;

        PLA_Gemm(pla_transa, pla_transb, Alpha_Mscalar, a, b, Beta_Mscalar, c);
        PLA_Set_check_parameters(PLA_NO_CHECK_PARAMETER);
	PLA_Obj_free( &Alpha_Mscalar);
	PLA_Obj_free( &Beta_Mscalar);
        return(0);
}

int  PLA_Gemm (int transa, int transb,  PLA_Obj alpha,
					PLA_Obj a,
					PLA_Obj b,
				 	PLA_Obj beta,
					PLA_Obj c )
{
MPI_Comm PLA_row, PLA_col;
int      PLA_nprows, PLA_npcols;
/* ALG_NB_SIZE is the blocking size that makes it run as fast as
	possible.  For right now it just takes this to be the blocking
	size of the inducing vector -- i.e. the blocking size of the
	matrices. */
	int ALG_NB_SIZE, nb, m, n, k, idummy;
	PLA_Template C_Template = NULL; /* A template is needed to get nb    */
	                              /* since nb is used as the ALG_NB_SIZE */
	                             /* C's blocking size chosen arbitrarily */
	MPI_Datatype   Cs_datatype;
        PLA_Gemm_enter(transa, transb, alpha, a, b, beta, c);
	PLA_Obj_template(c, &C_Template);
        PLA_Temp_comm_row_info(C_Template, &PLA_row, &idummy, &PLA_npcols);
        PLA_Temp_comm_col_info(C_Template, &PLA_col, &idummy, &PLA_nprows);
 	PLA_Temp_nb(C_Template, &nb);   
 	PLA_Environ_nb_alg(PLA_OP_PAN_PAN, C_Template, &ALG_NB_SIZE);   
	PLA_Obj_datatype(c, &Cs_datatype);
	if((transa == PLA_NO_TRANSPOSE) && (transb == PLA_NO_TRANSPOSE))
        {
           PLA_Obj_global_length(a, &m);
           PLA_Obj_global_width(a, &k);
           PLA_Obj_global_width(b, &n);
           if(  ((m*10) < k) && ((n*10) < k) && (m < (nb * PLA_nprows /2)) &&
                                                (n < (nb * PLA_nprows /2)) )
              PLA_Gemm_notransa_notransb_4(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if ( ( ((n*5) < k) && ((n*5) < m) ) ||
                     ( ((m*5) < k) && ((n*5) < k) ) )
              PLA_Gemm_notransa_notransb_2(ALG_NB_SIZE,  transa, transb,alpha, a, b,
                                                        beta, c);
           else if ( ((m*5) < k) && ((m*5) < n) )
              PLA_Gemm_notransa_notransb_3(ALG_NB_SIZE,  transa, transb,alpha, a, b,
                                                        beta, c);
           else
              PLA_Gemm_notransa_notransb_1(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
        } /* NT and TN cases not yet set to optimize shape results */
        else if((transa == PLA_NO_TRANSPOSE) && 
                ((transb == PLA_TRANSPOSE) || (transb == PLA_CONJ_TRANS)))
        {
           PLA_Obj_global_length(a, &m);
           PLA_Obj_global_width(a, &k);
           PLA_Obj_global_length(b, &n);
           if(  ((m*10) < k) && ((n*10) < k) && (m < (nb * PLA_nprows /2)) &&
                                                (n < (nb * PLA_nprows /2)) )
              PLA_Gemm_notransa_transb_4(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if ( ( ((k*5) < n) && ((k*5) < m) ) ||
                     ( ((k*5) < m) && ((n*5) < m) ) )
              PLA_Gemm_notransa_transb_2(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if ( ((m*5) < k) && ((m*5) < n) )
              PLA_Gemm_notransa_transb_3(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else
              PLA_Gemm_notransa_transb_1(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
        }
        else if( ((transa == PLA_TRANSPOSE)  || (PLA_CONJ_TRANS == transa))
                && (transb == PLA_NO_TRANSPOSE))
        {
           PLA_Obj_global_width(a, &m);
           PLA_Obj_global_length(a, &k);
           PLA_Obj_global_width(b, &n);
           if(  ((m*10) < k) && ((n*10) < k) && (m < (nb * PLA_nprows /2)) &&
                                                (n < (nb * PLA_nprows /2)) )
              PLA_Gemm_transa_notransb_4(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if  (  ( ((k*5) < n) && ((k*5) < m) )  || 
                      ( ((k*5) < m) && ((n*5) < m) ) ) 
              PLA_Gemm_transa_notransb_2(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if  ( ((n*5) < m) && ((n*5) < k) )
              PLA_Gemm_transa_notransb_3(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else
              PLA_Gemm_transa_notransb_1(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
        }
	else if( ((transa == PLA_TRANSPOSE) || (PLA_CONJ_TRANS == transa))
              && ((transb == PLA_TRANSPOSE)|| (PLA_CONJ_TRANS == transb)) )
        {
           PLA_Obj_global_width(a, &m);
           PLA_Obj_global_length(a, &k);
           PLA_Obj_global_length(b, &n);
           if(  ((m*5) < k) && ((n*5) < k) && (m < (nb * PLA_nprows /2)) &&
                                                (n < (nb * PLA_nprows /2)) )
              PLA_Gemm_transa_transb_4(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if ( ( ((n*1) < k) && ((n*1) < m) ) ||
                     ( ((m*1) < k) && ((n*1) < k) ) )
              PLA_Gemm_transa_transb_2(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else if ( ( ((m*1) < k) && ((m*1) < n) ) ||
                     ( (m < n) && (k < n) ) )
              PLA_Gemm_transa_transb_3(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
           else
              PLA_Gemm_transa_transb_1(ALG_NB_SIZE, transa, transb, alpha, a, b,
                                                        beta, c);
        }
	else printf("Unhandled kind of PLA_Gemm call\n");
        PLA_Gemm_exit(transa, transb, alpha, a, b, beta, c);
        return(0);
}

