#include <stdio.h>
#include "mpi.h"
#include "mmpi.h"
#include "SL.h"
#include "MISC_DEFS.h"
#include "PLA.MMPI.h"
#include "PLA_LAObj_mmpi_consumers.h"

int PLA0_LAObj_write_consumer( SL_LAObj laobj, 
                               int source,
                               MMPI_Buf buffer )
/* 
 *  PLA_LAObj_fill_consumer
 *
 *  Purpose: receive message and add it to appropriate part of laobj
 */
{
  MPI_Datatype        datatype;
  int                 typesize, 
                      current_pos, 
                      size;
  SL_LAObj_obj_type   objtype;
  void               *temp_buffer, *mp, *mpp;
  SL_LAObj            cur    = NULL;
  int                 curpos_row, curpos_col, 
                      curcol, 
                      size_top, size_left,
                      i;

  /* get object type */
  SL_LAObj_objtype( laobj, &objtype );

  /* get data type information */
  SL_LAObj_datatype( laobj, &datatype );
  MPI_Type_size( datatype, &typesize );


/*printf("enter write consumer\n"); */

  if ( objtype == SL_VECTOR ) {
    /* get location of where to put incoming data */
    MMPI_Buf_unpack( buffer, &current_pos, 1, MPI_INT );
    /* get size of incoming data */
    MMPI_Buf_unpack( buffer, &size, 1, MPI_INT );
    /* get incoming data */
    temp_buffer = (void *) malloc( size * typesize );
    MMPI_Buf_unpack( buffer, temp_buffer, size, datatype );
    /* take view of appropriate part of global object */
    SL_Vector_subvector( laobj, size, current_pos, &cur );
    /* copy incoming data into appropriate portion of global object */
/*    PLA0_LAObj_set_local_contents( SL_NOTRANS, size, 1, temp_buffer, 1,
			            1, cur ); */
    PLA0_LAObj_set_local_contents1( SL_NOTRANS, datatype, temp_buffer, 1,
                                    1, cur );
    /* free temporary buffer */
    free( temp_buffer );
    /* free view */
    SL_LAObj_free( &cur );
  }
  else if ( objtype == SL_MATRIX ) {
    /* get location of where to put incoming data */
    MMPI_Buf_unpack( buffer, &curpos_row, 1, MPI_INT );
    MMPI_Buf_unpack( buffer, &curpos_col, 1, MPI_INT );
    /* get size of incoming data */
    MMPI_Buf_unpack( buffer, &size_top, 1, MPI_INT );
    MMPI_Buf_unpack( buffer, &size_left, 1, MPI_INT );
    /* get incoming data */
    mpp = mp = (void *) malloc( size_top * size_left * typesize );
    for ( i=0; i<size_left; i++ ) {
      MMPI_Buf_unpack( buffer, mpp, size_top, datatype );
      /* mpp += size_top*typesize; */
      mpp = (void*)(((char*)mpp) + size_top*typesize);
    }
    /* take view of appropriate part of global object */
    SL_Matrix_submatrix( laobj, size_top, size_left, 
			 curpos_row, curpos_col, &cur );
    /* copy incoming data into appropriate portion of global object */
 /*   PLA0_LAObj_set_local_contents( SL_NOTRANS, size_top, size_left,
				   mp, size_top, 1, cur ); */
    PLA0_LAObj_set_local_contents1( SL_NOTRANS, datatype, 
                                   mp, size_top, 1, cur );
      /* free temporary buffer */
    free( mp );
    /* free view */
    SL_LAObj_free( &cur );
  }
    
/*printf("exit write consumer\n"); */

  return MPI_SUCCESS;
}


int PLA0_LAObj_read_consumer( SL_LAObj laobj, 
                              int source,
                              MMPI_Buf buffer )
/* 
 *  PLA_LAObj_read_consumer
 *
 *  Purpose: receives message that returns desired data from
 *           another processor
 */
{
  MPI_Datatype        datatype;
  int                 typesize, 
                      size;
  SL_LAObj_obj_type   objtype;
  void               *mp, *mpp, *vp;
  int                 size_top, size_left,
                      i, 
                      mp_ld;

  SL_LAObj_objtype( laobj, &objtype );

  /* get data type information */
  SL_LAObj_datatype( laobj, &datatype );
  MPI_Type_size( datatype, &typesize );

/*printf("enter read consumer message from %d\n", source);
fflush( stdout ); */
  if ( objtype == SL_VECTOR ) {
    /* get location of where to put incoming data */
    MMPI_Buf_unpack( buffer, &vp, sizeof( void * ), MPI_CHAR );
    /* get size of incoming data */
    MMPI_Buf_unpack( buffer, &size, 1, MPI_INT );
    /* get incoming data */
    MMPI_Buf_unpack( buffer, vp, size, datatype );
  }
  else if ( objtype == SL_MATRIX ) {
    /* get location of where to put incoming data and leading dimension */
    MMPI_Buf_unpack( buffer, &mp, sizeof( void *), MPI_CHAR );
    MMPI_Buf_unpack( buffer, &mp_ld, 1, MPI_INT );
    /* get size of incoming data */
    MMPI_Buf_unpack( buffer, &size_top, 1, MPI_INT );
    MMPI_Buf_unpack( buffer, &size_left, 1, MPI_INT );
    /* get incoming data */
    mpp = mp;
    for ( i=0; i<size_left; i++ ) {
      MMPI_Buf_unpack( buffer, mpp, size_top, datatype );
      /* mpp += mp_ld*typesize; */
      mpp = (void*) (((char*) mpp) + mp_ld*typesize);
    }
  }
    
/* printf("exit read consumer\n"); */

  return MPI_SUCCESS;
}


int PLA0_LAObj_read_bouncer( SL_LAObj laobj, 
                             int source,
                             MMPI_Buf buffer )
/* 
 *  PLA_LAObj_fill_consumer
 *
 *  Purpose: receive message and add it to appropriate part of laobj
 */
{
  MPI_Datatype        datatype;
  int                 typesize, 
                      current_pos, 
                      size;
  SL_LAObj_obj_type   objtype;
  void               *temp_buffer, *mp, *mpp, *vp_return,
                     *mp_return;
  SL_LAObj            cur    = NULL;
  int                 curpos_row, curpos_col, 
                      curcol, 
                      size_top, size_left,
                      i, laobj_ld, mp_ld;
  void               *laobj_buf;
  struct PLA0_Fill_info *laobj_fill;

/*printf("enter read bouncer\n"); */

  laobj_fill = ( struct PLA0_Fill_info * ) laobj->fill;

  /* get object type */
  SL_LAObj_objtype( laobj, &objtype );

  /* get data type information */
  SL_LAObj_datatype( laobj, &datatype );
  MPI_Type_size( datatype, &typesize );


  if ( objtype == SL_VECTOR ) {
    /* get location of where to get requested data */
    MMPI_Buf_unpack( buffer, &current_pos, 1, MPI_INT );
    /* get size of requested data */
    MMPI_Buf_unpack( buffer, &size, 1, MPI_INT );
    /* get bounce information */
    MMPI_Buf_unpack( buffer, &vp_return, sizeof( void * ), MPI_CHAR );
    /* take view of appropriate part of global object */
    SL_Vector_subvector( laobj, size, current_pos, &cur );
    SL_LAObj_buffer( laobj, &laobj_buf );
    /* initiate return message */
    MMPI_Con_init( laobj_fill->consumer_receive, &laobj_fill->buffer );
    /* pack return data */
    MMPI_Buf_pack( &vp_return, sizeof( void * ), MPI_CHAR, &laobj_fill->buffer );
    MMPI_Buf_pack( &size, 1, MPI_INT, &laobj_fill->buffer );
    MMPI_Buf_pack( laobj_buf, size, datatype, &laobj_fill->buffer );
    /* send return message */
    MMPI_Con_send( laobj_fill->buffer, source, laobj_fill->consumer_receive );
    /* free view */
    SL_LAObj_free( &cur );
  }
  else if ( objtype == SL_MATRIX ) {
    SL_LAObj_local_lld( laobj, &laobj_ld );
    /* get location of where to get requested data */
    MMPI_Buf_unpack( buffer, &curpos_row, 1, MPI_INT );
    MMPI_Buf_unpack( buffer, &curpos_col, 1, MPI_INT );
    /* get size of requested data */
    MMPI_Buf_unpack( buffer, &size_top, 1, MPI_INT );
    MMPI_Buf_unpack( buffer, &size_left, 1, MPI_INT );
    /* get bounce information */
    MMPI_Buf_unpack( buffer, &mp_return, sizeof( void * ), MPI_CHAR );
    MMPI_Buf_unpack( buffer, &mp_ld, 1, MPI_INT );
    /* take view of appropriate part of global object */
    SL_Matrix_submatrix( laobj, size_top, size_left, curpos_row, curpos_col, 
                         &cur );
    SL_LAObj_buffer( cur, &mpp );
    /* initiate return message */
    MMPI_Con_init( laobj_fill->consumer_receive, &laobj_fill->buffer );
    /* pack return data */
    MMPI_Buf_pack( &mp_return, sizeof( void * ), MPI_CHAR, &laobj_fill->buffer );
    MMPI_Buf_pack( &mp_ld, 1, MPI_INT, &laobj_fill->buffer );
    MMPI_Buf_pack( &size_top, 1, MPI_INT, &laobj_fill->buffer );
    MMPI_Buf_pack( &size_left, 1, MPI_INT, &laobj_fill->buffer );
    for (i=0; i<size_left; i++) {
      MMPI_Buf_pack( mpp, size_top, datatype, &laobj_fill->buffer );
      /* mpp += laobj_ld*typesize; */
      mpp = (void*) (((char*) mpp) + laobj_ld*typesize);
    } 
    /* send return message */
    MMPI_Con_send( laobj_fill->buffer, source, laobj_fill->consumer_receive );
    /* free view */
    SL_LAObj_free( &cur );
  } 
    
/* printf("exit read bouncer\n"); */

  return MPI_SUCCESS;
}
