#include "SL.h"
#include <stdio.h>
#include "PLA_to_from_GA.h"

#define PLA0_READ   1001
#define PLA0_WRITE  1002
typedef int           *GA_Object;

/********************************************************************/
int PLA0_GA_to_PLA_Create(GA_Object    ga_object,
                          SL_LAObj    *pla_object)
/********************************************************************/
/*                                                                  */
/* This routine queries the GA object and creates the corresponding */
/* PLA object.  It does not copy data between the two objects       */
/* Note : Really the only kind of PLA_LAObj that can be created is  */
/*        a matrix as the GA library does not include distributed   */
/*        1 and 1.5-D objects (Vectors and Multivectors).  Perhaps  */
/*        this routine could be expanded to create Pvectors and     */
/*        Pvectors where appropriate however.                       */
/* Further note : The template must be passed in because the GA and */
/* SL/PLA routines may not use the same grid.                       */
/*                                                                  */
/********************************************************************/
{
   int return_value = 0; /* an error code */

   int                 GA_Datatype, GA_length, GA_width, PLA_Datatype;
   SL_Template         pla_template = NULL;
   SL_LAObj            pla_obj_2 = NULL;
   MPI_Comm            comm;

   pla_object = &pla_obj_2; /* Temporary patch */
   ga_inquire_(ga_object, &GA_Datatype, &GA_length, &GA_width);
/*Right now -- just testing for creation with GA_Create -- not GA_Create_irregular*/
/*so I assume that the offsets are 0. Actually, there don't seem to be inquiry    */
/*routines for all of the stuff in the Irreg object anyway -- I have to check on  */
/* this!!!!!!!!!!!!!!!!!                                                          */

/* NOTE : IMHO the name of the Temp routines should be expanded to Template   */
/*        It may be a complete pain in the neck but Temp is just too evocative*/
/*        of "temporary".                                                     */
   PLA_Datatype = MPI_DOUBLE;
   fprintf(stdout, "About to call SL_Temp2\n");
   fflush(stdout);
   SL_Temp2(&pla_template);
   fprintf(stdout, "Just called SL_Temp2\n");
   fflush(stdout);
   return_value += PLA0_Matrix_create(PLA_Datatype, GA_length, GA_width,
                              pla_template, 0, 0, pla_object);
   return(return_value);
}

int PLA0_GA_to_PLA_Copy(  GA_Object    ga_object,
                          SL_LAObj    *pla_object)
{
   int return_value = 0;
   int GA_length, GA_width, i, j;
   int GA_Datatype;
   int GA_myProcessor;
   int row_low, row_hi, col_low, col_hi;
   int ga_index, ga_lld;

   PLA0_LAObj_open(PLA0_WRITE, *pla_object);

   return_value += ga_inquire_(ga_object, &GA_Datatype, &GA_length, &GA_width);
/* We don't handle (for now) as there appears to be no need?  Copying       */
/* between objects of different datatypes -- so the datatype can be gleaned */
/* from the destination object.  Might add check later.                     */
   ga_distribution_(ga_object, &GA_myProcessor,&row_low,&row_hi,&col_low,&col_hi);

/* GA has a GA_ACC and a GA_ACCESS -- the two have nothing in common so don't */
/* confuse them.                                                              */
   ga_access_(ga_object, &row_low,&row_hi,&col_low,&col_hi,&ga_index, &ga_lld);

   return_value += PLA0_LAObj_set_matrix_in_global_obj(*pla_object,
       (row_hi-row_low+1), (col_hi-col_low+1), row_low-1, col_low-1,
       (void*) ga_index, ga_lld);

   ga_release_(ga_object, &row_low, &row_hi, &col_low, &col_hi);

/* Final sync */
   return_value += PLA0_LAObj_close(*pla_object);
   return(return_value);
}

int PLA0_PLA_to_GA_Copy(  SL_LAObj    *pla_object,
                          GA_Object    ga_object)
{
   int return_value = 0;
   int GA_length, GA_width, i, j;
   int GA_Datatype;
   int GA_myProcessor;
   int row_low, row_hi, col_low, col_hi;
   int ga_index, ga_lld;

   PLA0_LAObj_open(PLA0_READ, *pla_object);

   return_value += ga_inquire_(ga_object, &GA_Datatype, &GA_length, &GA_width);
/* We don't handle (for now) as there appears to be no need?  Copying       */
/* between objects of different datatypes -- so the datatype can be gleaned */
/* from the destination object.  Might add check later.                     */
   ga_distribution_(ga_object, &GA_myProcessor,&row_low,&row_hi,&col_low,&col_hi);

/* GA has a GA_ACC and a GA_ACCESS -- the two have nothing in common so don't */
/* confuse them.                                                              */
   ga_access_(ga_object, &row_low,&row_hi,&col_low,&col_hi,&ga_index, &ga_lld);

   return_value += PLA0_LAObj_get_matrix_from_global_obj(*pla_object,
       (row_hi-row_low+1), (col_hi-col_low+1), row_low-1, col_low-1,
       (void*)ga_index, ga_lld);

   ga_release_(ga_object, &row_low, &row_hi, &col_low, &col_hi);

/* Final sync */
   return_value += PLA0_LAObj_close(*pla_object);
   return(return_value);
}



 
int SL_Temp2(SL_Template *template)
 
{
  int  nb;              /* Scalable BLAS blocking parameter        */
  int  r, c;            /* logical mesh dimensions                 */
  int  me;              /* my node index                           */
  int  ndims, dims[2];  /* for 2-D MPI_Comm creation               */
  int  periods[2];      /* S.A.A. Do we wrap the Comm. mesh        */
  int  reorder;         /* Allow/disallow re-ordering of proc #s   */
  MPI_Comm comm2d;      /* The 2-D communicator used               */
  SL_Template Second_Template = NULL;
  int dummy_argc = 2;
  char dummy_argv[2][6]; 
  int hello;

  template = &Second_Template;
		 /* Initialize MPI                                  */
		 /* Extract me                                      */
  fprintf(stdout, "Started SL_Temp2 -- temporary\n");
  fflush(stdout);
/*  strcpy(dummy_argv[0],"mpirun");
  strcpy(dummy_argv[1], "test");
  fscanf(stdin,"%d", &hello); 
  MPI_Init(&dummy_argc, &dummy_argv);  */
  fprintf(stdout, "Continue 1 SL_Temp2 -- temporary\n");
  fflush(stdout);
  MPI_Comm_rank( MPI_COMM_WORLD, &me );
  fprintf(stdout, "Continue 2 SL_Temp2 -- temporary\n");
  fflush(stdout);

  if ( me == 0 )
  {
	 printf("Input #procs/row, #procs/col\n");
	 scanf( "%d%d", &r, &c );
  }
  MPI_Bcast( &r, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &c, 1, MPI_INT, 0, MPI_COMM_WORLD ); 
  /* Create 2D communicator    -- make a function                  */
  ndims = 2;                         /* 2-Dimensional Communicator */
  dims[0] = r; dims[1] = c;                /* r x c grid */
  periods[0] = TRUE; periods[1] = TRUE;    /* Torus */
  reorder = FALSE;       /* dis-allows processors to be re-ordered */
/*  MPI_Cart_create(MPI_COMM_WORLD, ndims, dims, periods, reorder, &comm2d); */
					 /* Get blocking parameter, nb, and broadcast       */
/*  if ( me == 0 )
  {
	 printf("Input nb\n");
	 scanf( "%d", &nb );
  }
  MPI_Bcast( &nb, 1, MPI_INT, 0, MPI_COMM_WORLD ); */
		 /* Initialize template                             */
/*  SL_Temp_create(comm2d, 10, 0, template);  */
}
 
