
static int PLA_SEQUENTIAL_CHECK_FLAG;
static int PLA_INSTRUMENT_FLAG;
static int PLA_CHECK_PARAMETER_FLAG;

/**************** START OF CORE ROUTINES **************************************/
void pla_Log_abort(
  const char * const who ,   /* function */
  const char * const file,
  const int          line,
  const char * const msg ,
  int error,
  int original_error );

void pla_Log_error(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg,
   int          this_error,
   int          original_error
   );

void pla_Log_message(
   const char * const who ,
   const char * const file,
   const int          line,
   const char * const msg
   );
/**************** END OF CORE ROUTINES ****************************************/
/*************  START "STANDARD" ERROR TESTS **********************************/
int pla_Obj_object_valid(
   PLA_Obj obj,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error);

int pla_Obj_objtype_valid(
   PLA_Obj object,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_objtype_equal(
   PLA_Obj obj,
   int object_type,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_datatype_valid(
   PLA_Obj object,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_datatype_equal(
   PLA_Obj object,
   MPI_Datatype datatype,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_template_equal(
   PLA_Obj object,
   PLA_Template template,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_equal (
   PLA_Obj object,
   int size,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_length_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_global_width_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_length_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_length_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

int pla_Obj_local_width_width_equal (
   PLA_Obj obj1,
   PLA_Obj obj2,
   const char * const file,
   const int          line,
   const char * const message,
   int original_error );

/*************  END "STANDARD" ERROR TESTS ************************************/

/************* START ENCODED --> STRING CONVERSIONS  **************************/
int pla_code_string_objtype(
   int Object_type,
   char *object_name);

int pla_code_string_datatype(
   MPI_Datatype Datatype,
   char *data_name);

void pla_errorcode_string(
   int error_code,
   char *error_code_string);

/************* END ENCODED --> STRING CONVERSIONS  ****************************/

/**************** START FLAG INQUIRY ROUTINES *********************************/
int   PLA_Get_check_sequential (int *check_sequential );
int   PLA_Set_check_sequential (int sequential_set );
int   PLA_Get_instrument (int *instrument );
int   PLA_Set_instrument (int instrument_set );
int   PLA_Get_check_parameters ( int *check_parameter );
int   PLA_Set_check_parameters ( int check_set );
/**************** END FLAG INQUIRY ROUTINES ***********************************/

/**************** START TEMPORARY COPY ROUTINE ********************************/
/* Copy a matrix to an mscalar */

int PLA_Copy_temp(PLA_Obj the_matrix, PLA_Obj the_mscalar);

/**************** END TEMPORARY COPY ROUTINE ********************************/

/**************** START OF STACK STUFF ****************************************/
struct Stack_node_base {
   char *Message;
   struct Stack_node_base *previous_frame;
   struct Stack_node_base *next_frame;
};

typedef struct Stack_node_base PLA_Stack_node;
typedef PLA_Stack_node *PLA_Stack_frame_obj;


/* Some globals -- we need them global */
static PLA_Stack_frame_obj PLA_TOP_OF_MESSAGE_STACK = NULL;    /* could do these by sets */
static PLA_Stack_frame_obj PLA_BOTTOM_OF_MESSAGE_STACK = NULL; /* in PLA_Init()          */


int PLA_Stack_get_tos( PLA_Stack_frame_obj *top_of_stack);
int PLA_Stack_set_tos( PLA_Stack_frame_obj  top_of_stack);
int PLA_Stack_push_current(char *Stack_string);
int PLA_Stack_pop_current(char *Stack_string);

/**************** END OF STACK STUFF ******************************************/

/******************************************************************************/
/******************************************************************************/
/*   END OF HEADERS ***********************************************************/
/*   START OF CODE  ***********************************************************/
/******************************************************************************/
/******************************************************************************/


/**************** START OF CORE ROUTINES **************************************/
#define PLA_Log_abort(subroutine, message, error, original_error)             \
   pla_Log_abort (subroutine, __FILE__, __LINE__, message, error, original_error)

#define PLA_Log_error(subroutine, message, error, total_error)              \
   pla_Log_error (subroutine, __FILE__, __LINE__, message, error, total_error)

#define PLA_Log_message(subroutine, message)                                \
   pla_Log_message (subroutine, __FILE__, __LINE__, message)

/**************** END OF CORE ROUTINES ****************************************/

/*************  START "STANDARD" ERROR TESTS **********************************/
#define PLA_Obj_object_valid( object, message, error, original_error ) \
do       \
if(error = pla_Obj_object_valid(object, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
}  while (0)

#define PLA_Obj_objtype_valid( object, message, error, original_error )      \
do                                                                     \
if(error = pla_Obj_objtype_valid(object, __FILE__, __LINE__, message, original_error))           \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_objtype_equal( object, object_type, message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_objtype_equal(object, object_type, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_datatype_valid( object, message, error, original_error )      \
do                                                                     \
if(error = pla_Obj_datatype_valid(object, __FILE__, __LINE__, message, original_error))           \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)


#define PLA_Obj_datatype_equal( object, datatype, message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_datatype_equal(object, datatype, __FILE__, __LINE__, message, original_error))       \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_template_equal( object, template, message, error, original_error )  \
do                                                                     \
if(error = pla_Obj_template_equal(object, template, __FILE__, __LINE__, message, original_error))        \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                              \
   if(0 == original_error) original_error = error;                             \
} while (0)

#define PLA_Obj_global_width_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                              \
   if(0 == original_error) original_error = error;                             \
} while (0)

#define PLA_Obj_local_length_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while(0)

#define PLA_Obj_local_width_equal( obj1, size , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_equal( obj1, size , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_length_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_length_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_width_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_global_width_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_global_width_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))    \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_length_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_length_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_length_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_width_length_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_length_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))     \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
} while (0)

#define PLA_Obj_local_width_width_equal( obj1, obj2 , message, error, original_error ) \
do                                                                     \
if(error = pla_Obj_local_width_width_equal( obj1, obj2 , __FILE__, __LINE__, message, original_error ))    \
{                                                                             \
   if(0 == original_error) original_error = error;                            \
}while(0);
/*************  END "STANDARD" ERROR TESTS ************************************/


/************* START ENCODED --> STRING CONVERSIONS  **************************/
/************* END ENCODED --> STRING CONVERSIONS  ****************************/

/**************** START FLAG INQUIRY ROUTINES *********************************/

/**************** END FLAG INQUIRY ROUTINES ***********************************/

/**************** START TEMPORARY COPY ROUTINE ********************************/
/* Copy a matrix to an mscalar */

/**************** END TEMPORARY COPY ROUTINE **********************************/


/**************** START STACK STUFF  ******************************************/
/**************** END OF STACK STUFF *****************************************/

