/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu

***************************************************************************/
#include "PLA.h"

int PLA_Conjugate( PLA_Obj x )
{
	int original_error = 0;


	int error_value, i;
	int object_type_x;
	int local_length_x, local_width_x, ld_x;
	int ld_x_buf, stride_x, orientation_x;

	MPI_Datatype datatype_x;
        PLA_Template template;
	PLA_Obj Neg_one = NULL;

	void *x_buffer, *x_buffer_cur;
	void *Neg_one_buffer;

	PLA_Obj_local_length(x, &local_length_x);
	PLA_Obj_local_width (x, &local_width_x);
	PLA_Obj_datatype(x, &datatype_x);
        PLA_Obj_template(x, &template);
	if( (MPI_DOUBLE == datatype_x) || (MPI_DOUBLE_COMPLEX == datatype_x))
		PLA_Mscalar_create( MPI_DOUBLE, PLA_ALL_ROWS, PLA_ALL_COLS,
								  1, 1, template, &Neg_one );
	else PLA_Mscalar_create( MPI_FLOAT, PLA_ALL_ROWS, PLA_ALL_COLS,
								  1, 1, template, &Neg_one );
	PLA_Obj_set_to_minus_one(Neg_one);
	PLA_Obj_local_buffer(Neg_one, &Neg_one_buffer);
	PLA_Obj_local_ldim(x, &ld_x);
	PLA_Obj_objtype(x,  &object_type_x);
	PLA_Obj_local_buffer(x, &x_buffer);
		  PLA_Obj_local_stride(x, &stride_x);
	x_buffer_cur = x_buffer;
	if ( (MPI_DOUBLE == datatype_x) || (MPI_FLOAT == datatype_x))
		local_length_x = local_length_x/2; /* treat AS complex -- not negate */
	if(local_length_x * local_width_x > 0)
	{
		 if( (MPI_DOUBLE_COMPLEX == datatype_x) || (MPI_DOUBLE == datatype_x))
		 {
			 stride_x = stride_x * 2;
			 for(i = 0; i < local_width_x; i++)
			 {
				 PLA_dscal(&local_length_x,(double*)Neg_one_buffer,
							  (((double*)(x_buffer_cur))+1),&stride_x);
				 x_buffer_cur = (void*)(((double*)(x_buffer_cur))+ld_x*2);
			 }
		  }
		 if( (MPI_COMPLEX == datatype_x) || (MPI_FLOAT == datatype_x))
		 {
			 stride_x = stride_x * 2;
			 for(i = 0; i < local_width_x; i++)
			 {
				 PLA_sscal(&local_length_x,(float*)Neg_one_buffer,
							  (((float*)(x_buffer_cur))+1),&stride_x);
				 x_buffer_cur = (void*)(((float*)(x_buffer_cur))+ld_x*2);
			 }
		  }
	}
	PLA_Obj_free(&Neg_one);
	return(PLA_SUCCESS);
}
