#include "PLA.h"
/* These need to go in MISC_DEFS.h */

int PLA_Local_symv(int uplo, PLA_Obj alpha, PLA_Obj a, PLA_Obj x,
                   PLA_Obj beta, PLA_Obj y)
{
   char uplo_char;
   int  error_level = 0;
   int  incx, incy;
   int  lda, local_width_a, local_length_a, n;
   int  local_width_x, local_length_x;
   int  local_width_y, local_length_y;
   int  local_width_alpha, local_length_alpha;
   int  local_width_beta, local_length_beta;
   MPI_Datatype datatype_x, datatype_y, datatype_a, datatype_alpha, datatype_beta;
   void  *alpha_buffer, *beta_buffer, *a_buffer, *x_buffer, *y_buffer;

   if(PLA_UPPER_TRIANGULAR == uplo) uplo_char = 'U';
   else uplo_char = 'L';

   PLA_Obj_local_length(a, &local_length_a);
   PLA_Obj_local_width(a, &local_width_a);
   PLA_Obj_local_ldim(a, &lda);
   PLA_Obj_local_buffer(a, &a_buffer);
   PLA_Obj_datatype(a, &datatype_a);

   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width(x, &local_width_x);
   PLA_Obj_local_stride(x, &incx);
   PLA_Obj_local_buffer(x, &x_buffer);
   PLA_Obj_datatype(x, &datatype_x);

   PLA_Obj_local_length(y, &local_length_y);
   PLA_Obj_local_width(y, &local_width_y);
   PLA_Obj_local_stride(y, &incy);
   PLA_Obj_local_buffer(y, &y_buffer);
   PLA_Obj_datatype(y, &datatype_y);

   PLA_Obj_local_buffer(alpha, &alpha_buffer);
   PLA_Obj_local_length(alpha, &local_length_alpha);
   PLA_Obj_local_width(alpha, &local_width_alpha);
   PLA_Obj_datatype(alpha, &datatype_alpha);

   PLA_Obj_local_buffer(beta, &beta_buffer);
   PLA_Obj_local_length(beta, &local_length_beta);
   PLA_Obj_local_width(beta, &local_width_beta);
   PLA_Obj_datatype(beta, &datatype_beta);
/* This seems like a LOT!!! of error checking, probably too much. */

   if( (datatype_a != datatype_x) || (datatype_x != datatype_y) ||
       (datatype_y != datatype_alpha) || (datatype_alpha != datatype_beta) ||
       (local_width_x != 1) || (local_width_y != 1) ||
       (local_width_alpha != 1) || (local_length_alpha != 1) ||
       (local_width_beta != 1) || (local_length_beta != 1) ||
       (local_length_a != local_width_a) || (local_width_a != local_length_x) ||
       (local_length_x != local_length_y) )
   {
      error_level = PLA_ERROR;
      return(error_level);
   }
   else if(local_length_a * local_width_a > 0)
   {
   	if(MPI_DOUBLE == datatype_a)
      	PLA_dsymv(&uplo_char, &local_length_a, (double*)alpha_buffer,
                   (double*)a_buffer, &lda, (double*)x_buffer, &incx,
                   (double*)beta_buffer, (double*)y_buffer, &incy);
      else if(MPI_FLOAT == datatype_a)
      	PLA_ssymv(&uplo_char, &local_length_a, (float*)alpha_buffer,
                   (float*)a_buffer, &lda, (float*)x_buffer, &incx,
                   (float*)beta_buffer, (float*)y_buffer, &incy);
   }

   return(error_level);
}

int PLA_Local_trmv(int uplo, int trans, int diag, PLA_Obj a, PLA_Obj x)
{
   char uplo_char, trans_char, diag_char;
   int  error_level = 0;
   int  incx;
   int  lda, local_width_a, local_length_a, n;
   int  local_width_x, local_length_x;
   MPI_Datatype datatype_x, datatype_a;
   void  *a_buffer, *x_buffer;

   if(PLA_UPPER_TRIANGULAR == uplo) uplo_char = 'U';
   else uplo_char = 'L';

   if(PLA_NO_TRANS == trans) trans_char = 'N';
   else trans_char = 'T';

   if( PLA_UNIT_DIAG == diag) diag_char = 'U';
   else diag_char = 'N';

   PLA_Obj_local_length(a, &local_length_a);
   PLA_Obj_local_width(a, &local_width_a);
   PLA_Obj_local_ldim(a, &lda);
   PLA_Obj_local_buffer(a, &a_buffer);
   PLA_Obj_datatype(a, &datatype_a);

   PLA_Obj_local_length(x, &local_length_x);
   PLA_Obj_local_width(x, &local_width_x);
   PLA_Obj_local_stride(x, &incx);
   PLA_Obj_local_buffer(x, &x_buffer);
   PLA_Obj_datatype(x, &datatype_x);


   if( (datatype_a != datatype_x) ||
       (local_width_x != 1) ||
       (local_length_a != local_width_a) || (local_width_a != local_length_x))
   {
      error_level = PLA_ERROR;
      return(error_level);
   }
   else if(local_length_a * local_width_a > 0)
   {
   	if(MPI_DOUBLE == datatype_a)
      	PLA_dtrmv(&uplo_char, &trans_char, &diag_char, &local_length_a,
                   (double*)a_buffer, &lda, (double*)x_buffer, &incx);
      else if(MPI_FLOAT == datatype_a)
      	PLA_strmv(&uplo_char, &trans_char, &diag_char, &local_length_a,
                   (float*)a_buffer, &lda, (float*)x_buffer, &incx);
   }

   return(error_level);
}
