/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@cs.utexas.edu
 
***************************************************************************/
#include "PLA.h"
#include "mmpi.h"
/* SHOULD the buffer size be a long???? */
int PLA_Mscalar_buf_size(MPI_Datatype      datatype,
		      int               owner_row,
		      int               owner_col,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               *bytes_required)
{
   int local_length, local_width, data_size;
   int myrow, mycol;

   MPI_Type_size(datatype, &data_size);
  PLA_Temp_mycoords( template, &myrow, &mycol);

   if ( (owner_row == myrow || owner_row == PLA_ALL_ROWS ) &&
        (owner_col == mycol || owner_col == PLA_ALL_COLS ) )
   {

      /* Determine the local length; */
      local_length     = global_length;
      local_width      = global_width;
   }
   else
   {
      local_length = 0;
      local_width = 0;
   }
   *bytes_required = data_size * local_length * local_width;
   return(PLA_SUCCESS);
 
}
/* Buffer flags
1 : A buffer is passed for use but the object does not exist
0 : A buffer is not passed and one should be malloced
-1: A buffer is not passed and one should NOT be malloced
*/

int PLA_Mscalar_create (MPI_Datatype      datatype,
		      int               owner_row,
		      int               owner_col,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      PLA_Obj          *new_mscalar)
{
   void *data_buffer = NULL;
   int  buffer_flag =  0; /* indicates that no buffer is passed but one should be made */
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   int retval;
   retval = PLA_Mscalar_create_base(datatype, owner_row,owner_col,
	                          global_length, global_width, template,
			 	  buffer_flag, buffer_size, data_buffer,
				  new_mscalar);
   /* One of my new fields */
   (*new_mscalar)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_mscalar)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
   return(retval);
}

int PLA_Mscalar_create_buf (MPI_Datatype      datatype,
		      int               owner_row,
		      int               owner_col,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      int               buffer_size,
		      void             *data_buffer,
		      PLA_Obj          *new_mscalar)
{
   int  buffer_flag = 1; /* indicates that a buffer is passed for use */
   int  retval;
   retval = PLA_Mscalar_create_base(datatype, owner_row,owner_col,
	                             global_length, global_width, template,
								 buffer_flag, buffer_size, data_buffer,
								 new_mscalar);
   (*new_mscalar)->allocation_type = PLA_USER_BUFFER;    /* vs. PLA_HEAP */
   (*new_mscalar)->has_data_space  = PLA_HAS_SPACE;      /* vs. PLA_NO_SPACE */
   return(retval);
}

int PLA_Mscalar_create_no_buf (MPI_Datatype      datatype,
		      int               owner_row,
		      int               owner_col,
		      int               global_length,
		      int               global_width,
		      PLA_Template      template,
		      PLA_Obj          *new_mscalar)
{
   int  buffer_flag = -1; /* indicates that a buffer is passed for use */
   void *data_buffer = NULL;
   int  buffer_size =  0; /* unimportant in this case -- may be "rolled into" the flag */
   int  retval;
   retval = PLA_Mscalar_create_base(datatype, owner_row,owner_col,
	                             global_length, global_width, template,
								 buffer_flag, buffer_size, data_buffer,
								 new_mscalar);
   (*new_mscalar)->allocation_type = PLA_UNDEFINED;    /* not defined until space attached  */
   (*new_mscalar)->has_data_space  = PLA_NO_SPACE;    /* vs. PLA_HAS_SPACE */
   return(retval);
}

int PLA_Mscalar_attach_buf(PLA_Obj mscalar, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(mscalar, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = mscalar->local_length * mscalar->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Mscalar_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else mscalar->master_buffer   = data_buffer;
   mscalar->allocation_type = PLA_USER_SUPPLIED;
   mscalar->has_data_space  = PLA_HAS_SPACE;
   return(PLA_SUCCESS);
}

int PLA_Mscalar_attach(PLA_Obj mscalar)
{
   /* Need to modularize the mscalar create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(mscalar, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = mscalar->local_length * mscalar->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      mscalar->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (mscalar->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Mscalar_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else mscalar->master_buffer = NULL;

   mscalar->allocation_type    = PLA_HEAP;
   mscalar->has_data_space     = PLA_HAS_SPACE;
   return(PLA_SUCCESS);
}

int PLA_Mscalar_create_base (MPI_Datatype      datatype,
                       int              owner_row,
                       int              owner_col,
      		       int              global_length,
		       int              global_width,
		       PLA_Template     template,
		       int              buffer_flag,
		       int              external_buffer_size,
		       void            *external_buffer,
		       PLA_Obj         *new_mscalar)
{
  PLA_Obj       mscalar;
  int           buf_size;
  int           myrow, mycol;
  int           error=0;
  int           data_size;
  PLA_Start_num zero_or_one;

  int           index; /* for getting a free Obj_struct */
  int           local_buffer_size_required = -1;

  if(! PLA_Initialized_object_structs())
     PLA_Init_object_structs(PLA_DEFAULT_OBJ_COUNT); /* some reasonable number of objects */

  PLA_Obj_free(new_mscalar);
  
  index = PLA_Get_free_object_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated object structure in Mscalar creation\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/*printf("Mscalar : Index is %5d and place is %ld\n", index,
&PLA_GLOBAL_OBJ_STRUCT[index]);*/
  mscalar = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);

  error += PLA_Temp_zero_or_one (template, &zero_or_one);

  error += PLA_Obj_init (mscalar);
  mscalar->obj_type         = PLA_MSCALAR;
  mscalar->datatype         = datatype;
  mscalar->global_length    = mscalar->master_length = global_length;
  mscalar->global_width     = mscalar->master_width  = global_width;
  mscalar->local_stride     = 1;
  mscalar->template         = template;
  mscalar->global_align_row = mscalar->master_align_row = (int) zero_or_one;
  mscalar->global_align_col = mscalar->master_align_col = (int) zero_or_one;

  mscalar->master_buffer = NULL;
  mscalar->local_length  = 0;
  mscalar->local_width   = 0;
  mscalar->local_lld     = 0;

  buf_size = 0;
  data_size = 0;


  PLA_Temp_mycoords( template, &myrow, &mycol);

   if ( (owner_row == myrow || owner_row == PLA_ALL_ROWS ) &&
        (owner_col == mycol || owner_col == PLA_ALL_COLS ) )
   {

      /* Determine the local length; */
      mscalar->local_length     = global_length;
      mscalar->local_width      = global_width;
      mscalar->local_lld        = global_length;
      if(-1 == buffer_flag) /* Create an object with NO!!! buffer */
         buf_size = -1;
      else buf_size = mscalar->local_length * mscalar->local_width;
/*printf("Mscalar buf size is %5d\n", buf_size); */
      error = MPI_Type_size (datatype, &data_size);
      if (MPI_SUCCESS != error)
      {
          char buffer[MPI_MAX_ERROR_STRING];
          int resultlen;
          printf (" Error %d returned from MPI_Type_size\n",error);
          MPI_Error_string (error, buffer, &resultlen);
          puts (buffer);
      }

      if (!data_size)
      {
         data_size = 1;
         printf (" MPI_Type_size returned zero byte length for data element\n");
         error = 1;
      }

      if(0 == buffer_flag)
      {
         if ( 0 != buf_size )
         {
            mscalar->master_buffer = (void *) pla_calloc (buf_size,data_size);
            if (NULL == mscalar->master_buffer )
            {
               printf(" Could not allocate buffer for %d elements \n", buf_size);
               printf(" Of size %d bytes each in PLA_Mscalar_create\n",data_size);
               return(1);
            }
         }
         else mscalar->master_buffer = NULL;
      } 
      else if(1 == buffer_flag)
      {
         local_buffer_size_required = data_size * buf_size;
         if(external_buffer_size < local_buffer_size_required)
         {
            printf("PLA_Mscalar_create_buf was called with an attached buffer that was too small\n");
            return(1);
         }
         mscalar->master_buffer = external_buffer;
      }
   }

  mscalar->length_master_buffer_byte = buf_size*(int)data_size;
  mscalar->buffer = mscalar->master_buffer;
  index = PLA_Get_free_views_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated views structure in Mscalar creation\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
  mscalar->views  = (int *)(&PLA_GLOBAL_VIEWS_STRUCT[index]);
  mscalar->views[0]= 1;
  mscalar->api = NULL;


  /* Set the first and last block size in both dimentions */

  mscalar->owner_top_block     = owner_row;
  mscalar->length_top_block    = global_length;
  mscalar->owner_bottom_block  = owner_row;
  mscalar->length_bottom_block = global_length;
  mscalar->owner_left_block    = owner_col;
  mscalar->width_left_block    = global_width;
  mscalar->owner_right_block   = owner_col;
  mscalar->width_right_block   = global_width;
  *new_mscalar = mscalar;
  return(error);
}

int PLA_Mscalar_submscalar (PLA_Obj old_mscalar, 
			   int      global_length, 
			   int      global_width,
			   int      align_row,
			   int      align_col,
			   PLA_Obj *new_mscalar)
{
  int width_right, length_bottom;
  PLA_Obj mscalar = NULL;
  int index;
  if ( old_mscalar->obj_type != PLA_MSCALAR)
  {
    printf ("Invalid object passed to PLA_Mscalar_submscalar: NonMscalar\n");
    return (1);
  }  

  if ( * new_mscalar != NULL ) PLA_Obj_free (new_mscalar);
  index = PLA_Get_free_object_struct();
  if(-1 == index)
  {
     printf("Could not get an allocated object structure in Mscalar creation\n");
     printf("either the user did not set aside any object structures\n");
     printf("or they didn't set aside any and the system number is low\n");
     /* Note : in the future this will call for doing a mark-and-copy  */
     /*        type deal -- IF the user did NOT do their own set-aside */
     /*        and just FAILING if they did.                           */
  }
/* printf("SubMscalar : Index is %5d and place is %ld\n", index,
&PLA_GLOBAL_OBJ_STRUCT[index]); */
  mscalar = (PLA_Obj) (&PLA_GLOBAL_OBJ_STRUCT[index]);

  /* 
     By default copy all the fields in one blow.  Then modify
     the ones that need modifying.
     */
  * mscalar = * old_mscalar;

  if ( align_row == PLA_ALIGN_FIRST ) align_row = 0;
  if ( align_col == PLA_ALIGN_FIRST ) align_col = 0;
  if ( global_length == PLA_DIM_ALL ) global_length = mscalar->global_length;
  if ( global_width == PLA_DIM_ALL ) global_width = mscalar->global_width;

  width_right = global_width - mscalar->global_width + align_col;
  length_bottom = global_length - mscalar->global_length + align_row; 

  if ( align_row || align_col || width_right || length_bottom )
  {
/*     fprintf ( MMPI_Log_file (),
               "Calling view shift - %d %d %d %d : %d %d %d %d\n",
               global_length, global_width, align_row, align_col,
               align_row, align_col, width_right, length_bottom );
*/
     PLA_Mscalar_view_shift ( mscalar, align_row, align_col,
                                       width_right, length_bottom );
  }

  mscalar->views[0]++;
  * new_mscalar = mscalar;

  return ( PLA_SUCCESS );
}

int PLA_Mscalar_view_shift (PLA_Obj mscalar, 
			   int      length_top,
			   int      width_left,
			   int      width_right,
			   int      length_bottom)
{
  int offset, data_size, error=0,
      global_length, global_width, global_align_row, global_align_col,
      owner_row, owner_col, myrow, mycol;
  
  if ( mscalar->obj_type != PLA_MSCALAR) {
    printf ("Invalid object passed to PLA_Mscalar_view_shift: NonMscalar\n");
    return (1);
  }  

  global_align_row = mscalar->global_align_row + length_top;
  global_align_col = mscalar->global_align_col + width_left;
  global_length    = mscalar->global_length    + length_bottom - length_top;
  global_width     = mscalar->global_width     + width_right   - width_left;

   if ( global_length < 0 ||
        global_width < 0  ||
        global_align_row < mscalar->master_align_row  ||
        global_align_col < mscalar->master_align_col  ||
        global_align_row + global_length >
           mscalar->master_align_row + mscalar->master_length ||
        global_align_col + global_width >
           mscalar->master_align_col + mscalar->master_width )
   {

/* Greg Morrow, 05/07/97 : commenting out the error messages */
/*
      printf ( "Shift outside of bounds in PLA_Mscalar_shift\n" );
      fprintf ( MMPI_Log_file (),
                "Shift outside of bounds in PLA_Mscalar_shift\n" );
      fprintf ( MMPI_Log_file (),
                "Original size = (%d,%d, %d,%d)\n", mscalar->global_length,
                                                   mscalar->global_width,
                                                   mscalar->global_align_row,
                                                   mscalar->global_align_col );
      fprintf ( MMPI_Log_file (),
                "Master size = (%d,%d, %d,%d)\n", mscalar->master_length,
                                                 mscalar->master_width,
                                                 mscalar->master_align_row,
                                                 mscalar->master_align_col );
      fprintf ( MMPI_Log_file (),
                "Desired size = (%d,%d, %d,%d)\n", global_length,
                                                  global_width,
                                                  global_align_row,
                                                  global_align_col );
      error=1;  
*/
   }
   
  /*if ( error ) return ( error );*/

  mscalar->global_length = global_length;
  mscalar->global_width = global_width;
  mscalar->global_align_row = global_align_row;
  mscalar->global_align_col = global_align_col;

  PLA_Obj_owner_row ( mscalar, & owner_row );
  PLA_Obj_owner_col ( mscalar, & owner_col );
  {
     PLA_Template template = NULL;
     PLA_Obj_template ( mscalar, & template );
     PLA_Temp_comm_row_rank ( template, & mycol );
     PLA_Temp_comm_col_rank ( template, & myrow );
  }

  if ( ( owner_row == myrow || owner_row == PLA_ALL_ROWS ) &&
       ( owner_col == mycol || owner_col == PLA_ALL_COLS ) )
  {

     mscalar->local_length  = global_length;
     mscalar->local_width   = global_width;

     MPI_Type_size (mscalar->datatype, &data_size);

     offset  = mscalar->global_align_col - mscalar->master_align_col;
     offset *= mscalar->local_lld;
     offset += mscalar->global_align_row - mscalar->master_align_row;
     offset *= (int) data_size;
    
     mscalar->buffer = (void *) ( (char *)(mscalar->master_buffer) + offset );
    
  }
  return ( PLA_SUCCESS );
}

int PLA_Mscalar_scalar_double (PLA_Template  template,
			      PLA_Obj    *double_zero,
			      PLA_Obj    *double_one)
     
{
  int     error =0;
  double *dpointer;

  error += PLA_Mscalar_create (MPI_DOUBLE, 
                              PLA_ALL_ROWS, PLA_ALL_COLS,
			      1,
			      1,
			      template,
			      double_zero);
  dpointer = (double *) (*double_zero)->master_buffer;
  *dpointer= (double) 0.;
  
  error += PLA_Mscalar_create (MPI_DOUBLE, 
                              PLA_ALL_ROWS, PLA_ALL_COLS,
			      1,
			      1,
			      template,
		              double_one);
  dpointer = (double *) (*double_one)->master_buffer;
  *dpointer= (double) 1.;

}

int PLA_Mscalar_local_all (PLA_Obj          mscalar,
			  MPI_Datatype *    datatype,
			  int  *            local_length,
			  int  *            local_width,
			  void **           buffer,
			  int  *            local_lld)
{
  int error=0;

  error += PLA_Obj_datatype       (mscalar,  datatype);
  error += PLA_Obj_local_length   (mscalar,  local_length);
  error += PLA_Obj_local_width    (mscalar,  local_width);
  error += PLA_Obj_local_buffer         (mscalar,  buffer); 
  error += PLA_Obj_local_ldim      (mscalar,  local_lld); 
  return (error);
}


int PLA_Mscalar_global_all (PLA_Obj          mscalar,
			   PLA_Template *     template)
     
{
  int error=0;
  error += PLA_Obj_template      (mscalar,  template); 
  return (error);
}

  
     
