/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

***************************************************************************/
#include "PLA.h"
#include <stdio.h>
#include "MISC_DEFS.h"

/* SHOULD the buffer size be a long???? */
int PLA_Pvector_buf_size(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       PLA_Template       template,
                       int               global_align,
                       int             *bytes_required)
{
   return(PLA_Panel_buf_size(datatype, project_onto, owner,
                                global_length, 1, template, global_align, bytes_required));
}

int PLA_Pmvector_buf_size(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template       template,
                       int               global_align,
                       int             *bytes_required)
{
   return(PLA_Panel_buf_size(datatype, project_onto, owner,
                                global_length, global_width, template, global_align,
                                bytes_required) );
}

/* Buffer flags
1 : A buffer is passed for use but the object does not exist
0 : A buffer is not passed and one should be malloced
-1: A buffer is not passed and one should NOT be malloced
*/

int PLA_Pvector_create(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       PLA_Template       template,
                       int               global_align,
                       PLA_Obj         *new_proj_vector)
{
   int return_value = 0;
   void *data_buffer = NULL;
   int buffer_flag = 0;
   int buffer_size = 0;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_vector)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_proj_vector)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
   return(return_value);
}

int PLA_Pvector_create_buf(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template      template,
                       int               global_align,
			              int               buffer_size,
			              void             *data_buffer,
                       PLA_Obj          *new_proj_vector)
{
   int return_value = 0;
   int buffer_flag = 1;
   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_vector)->allocation_type = PLA_USER_SUPPLIED;
   (*new_proj_vector)->has_data_space  = PLA_HAS_SPACE;
   return(return_value);
}

int PLA_Pvector_create_no_buf(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       PLA_Template       template,
                       int               global_align,
                       PLA_Obj         *new_proj_vector)
{
   int return_value = 0;
   void *data_buffer = NULL;
   int buffer_flag = -1;
   int buffer_size = 0;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, 1, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_vector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_vector)->allocation_type = PLA_UNDEFINED;
   (*new_proj_vector)->has_data_space  = PLA_NO_SPACE;
   return(return_value);
}

int PLA_Pvector_attach_buf(PLA_Obj pvector, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(pvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = pvector->local_length * pvector->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Pvector_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else pvector->master_buffer   = data_buffer;
   pvector->allocation_type = PLA_USER_SUPPLIED;
   pvector->has_data_space  = PLA_HAS_SPACE;
}

int PLA_Pvector_attach(PLA_Obj pvector)
{
   /* Need to modularize the pvector create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(pvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = pvector->local_length * pvector->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      pvector->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (pvector->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Pvector_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else pvector->master_buffer = NULL;

   pvector->allocation_type    = PLA_HEAP;
   pvector->has_data_space     = PLA_HAS_SPACE;
}

/**************************  THE P____M!!!______VECTORS START HERE ***********/
int PLA_Pmvector_create(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template      template,
                       int               global_align,
                       PLA_Obj          *new_proj_mvector)
{
   int return_value = 0;
   void *data_buffer = NULL;
   int buffer_flag = 0;
   int buffer_size = 0;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_mvector)->allocation_type = PLA_HEAP;       /* vs. PLA_USER_BUFFER */
   (*new_proj_mvector)->has_data_space  = PLA_HAS_SPACE;  /* vs. PLA_NO_SPACE */
   return(return_value);
}

int PLA_Pmvector_create_buf(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template       template,
                       int               global_align,
			              int               buffer_size,
			              void             *data_buffer,
                       PLA_Obj         *new_proj_mvector)
{
   int return_value = 0;
   int buffer_flag = 1;
   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_mvector)->allocation_type = PLA_USER_SUPPLIED;
   (*new_proj_mvector)->has_data_space  = PLA_HAS_SPACE;
   return(return_value);
}

int PLA_Pmvector_create_no_buf(MPI_Datatype     datatype,
                       int               project_onto,
                       int               owner,
                       int               global_length,
                       int               global_width,
                       PLA_Template       template,
                       int               global_align,
                       PLA_Obj         *new_proj_mvector)
{
   int return_value = 0;
   void *data_buffer = NULL;
   int buffer_flag = -1;
   int buffer_size = 0;

   if(project_onto == PLA_PROJ_ONTO_ROW)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_ROW, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else if(project_onto == PLA_PROJ_ONTO_COL)
   {
      return_value += PLA_Panel_create_base(datatype, PLA_PROJ_ONTO_COL, owner,
                                      global_length, global_width, template, global_align,
                                      buffer_flag, buffer_size, data_buffer,
                                      new_proj_mvector);
   }
   else
   {
      return_value += UNKNOWN_PROJECTION;
      return(return_value);
   }       /* One of my new fields */
   (*new_proj_mvector)->allocation_type = PLA_UNDEFINED;
   (*new_proj_mvector)->has_data_space  = PLA_NO_SPACE;
   return(return_value);
}

int PLA_Pmvector_attach_buf(PLA_Obj pmvector, int buffer_size, void *data_buffer)
{
   int data_size, buf_size, local_buffer_size_required;
   MPI_Datatype datatype;

   PLA_Obj_datatype(pmvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = pmvector->local_length * pmvector->local_width;
   local_buffer_size_required = data_size * buf_size;
   if(buffer_size < local_buffer_size_required)
   {
      printf("PLA_Pmvector_attach_buf was called with an attach buffer that was too small\n");
      return(1);
   }
   else pmvector->master_buffer   = data_buffer;
   pmvector->allocation_type = PLA_USER_SUPPLIED;
   pmvector->has_data_space  = PLA_HAS_SPACE;
}

int PLA_Pmvector_attach(PLA_Obj pmvector)
{
   /* Need to modularize the pmvector create routine */
   /* This routine allocs buffer space for a passed in object -- actually,
      I am considering making them all just call PLA_Object_attach
      because they all do the same thing */
   int data_size, buf_size;
   MPI_Datatype datatype;

   PLA_Obj_datatype(pmvector, &datatype);
	/* Check for size correctness,
	   Set the flag that indicates that space is passed in */
   MPI_Type_size (datatype, &data_size);
   buf_size = pmvector->local_length * pmvector->local_width;
   if (0 != buf_size)
   {
      MPI_Type_size (datatype, &data_size);
      pmvector->master_buffer = (void *) pla_calloc (buf_size,(int)data_size);
      if (pmvector->master_buffer == NULL)
      {
         printf (" Could not allocate buffer of %d bytes in PLA_Pmvector_attach\n",
	      buf_size*(int)data_size);
         return(1);
      }
   }
   else pmvector->master_buffer = NULL;

   pmvector->allocation_type    = PLA_HEAP;
   pmvector->has_data_space     = PLA_HAS_SPACE;
}



