#include "PLA.h"

double CRAYNULLVALUE=1;
void * CRAYNULLPOINTER=&CRAYNULLVALUE;

#define BUFCHK(a) ((a)==NULL?CRAYNULLPOINTER:a)

int pla_Copy_pmv2mat ( PLA_Obj source, PLA_Obj destin )
{
   int p, owner, rank, size, proj_onto, block_length, offset,
       local_length, local_width, source_ldim, destin_ldim,
       global_length, global_width, global_align_row, global_align_col,
       * ldims = NULL, * counts = NULL, * displs = NULL;
   void * bsource = NULL, * bdestin = NULL,
        * bsend = NULL,   * brecv = NULL;
   MPI_Aint typesize;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   MPI_Comm comm = MPI_COMM_NULL;
   PLA_Template template = NULL;
   PLA_Obj view = NULL;

   PLA_Obj_template ( source, & template );
   PLA_Obj_datatype ( source, & datatype );
   MPI_Type_extent ( datatype, & typesize );
   PLA_Obj_project_onto ( source, & proj_onto );

   PLA_Obj_global_length ( destin, & global_length );
   PLA_Obj_global_width ( destin, & global_width );
   PLA_Obj_global_align_row ( destin, & global_align_row );
   PLA_Obj_global_align_col ( destin, & global_align_col );
   PLA_Obj_local_length ( destin, & local_length );
   PLA_Obj_local_width ( destin, & local_width );

   PLA_Obj_local_buffer ( source, & bsource );
   PLA_Obj_local_buffer ( destin, & bdestin );
   PLA_Obj_local_ldim ( source, & source_ldim );
   PLA_Obj_local_ldim ( destin, & destin_ldim );

   if ( proj_onto == PLA_PROJ_ONTO_COL )
   {
      PLA_Temp_comm_row_info ( template, & comm, & rank, & size );
      PLA_Obj_owner_col ( source, & owner );
   }
   else
   {
      PLA_Temp_comm_col_info ( template, & comm, & rank, & size );
      PLA_Obj_owner_row ( source, & owner );
   }


   ldims = pla_calloc ( size, sizeof ( int ) );
   counts = pla_calloc ( size, sizeof ( int ) );
   displs = pla_calloc ( size, sizeof ( int ) );

   for ( p = 0 ; p < size ; p++ )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         ldims [ p ] = local_length;
         PLA_Temp_matrix_distant_width ( template, p,
                                         global_width, global_align_col,
                                         & counts [ p ] );
         counts [ p ] *= local_length;
      }
      else
      {
         PLA_Temp_matrix_distant_length ( template, p,
                                          global_length, global_align_row,
                                          & ldims [ p ] );
         counts [ p ] = ldims [ p ] * local_width;
      }

      if ( p > 0 ) displs [ p ] = displs [ p - 1 ] + counts [ p - 1 ];
   }


/* pack */

   if ( local_length * local_width )
      brecv = pla_calloc ( local_length * local_width, typesize );

   if ( rank == owner )
   {
      for ( p = 0 ; p < size ; p++ )  counts [ p ] = 0;

      PLA_Obj_local_length ( source, & local_length );
      PLA_Obj_local_width ( source, & local_width );
      if ( local_length * local_width )
         bsend = pla_calloc ( local_length * local_width, typesize );

      if ( proj_onto == PLA_PROJ_ONTO_COL )
         PLA_Temp_matrix_first_col_block ( template, global_width, global_align_col,
                                           & p, & block_length );
      else
         PLA_Temp_matrix_first_row_block ( template, global_length, global_align_row,
                                           & p, & block_length );

      while ( block_length > 0 )
      {
         offset = ( displs [ p ] + counts [ p ] ) * typesize;

         if ( proj_onto == PLA_PROJ_ONTO_COL )
         {
            global_width -= block_length;
            global_align_col += block_length;

            copy_block ( local_length, block_length, typesize,
                         bsource, source_ldim,
                         ( char * ) bsend + offset, ldims [ p ] );

            counts [ p ] += block_length * ldims [ p ];
            bsource = ( char * ) bsource + source_ldim * block_length * typesize;

            PLA_Temp_matrix_first_col_block ( template, global_width, global_align_col,
                                              & p, & block_length );
        }
        else
        {
           global_length -= block_length;
           global_align_row += block_length;

           copy_block ( block_length, local_width, typesize,
                        bsource, source_ldim,
                        ( char * ) bsend + offset, ldims [ p ] );

           counts [ p ] += block_length;
           bsource = ( char * ) bsource + block_length * typesize;

           PLA_Temp_matrix_first_row_block ( template, global_length, global_align_row,
                                             & p, & block_length );
         }
      }

      if ( proj_onto == PLA_PROJ_ONTO_ROW )
         for ( p = 0 ; p < size ; p++ ) counts [ p ] *= local_width;
   }


/* pass message */

   MPI_Scatterv ( BUFCHK(bsend), counts, displs, datatype,
                  BUFCHK(brecv), counts [ rank ], datatype, owner, comm );


/* unpack */

   PLA_Obj_local_length ( destin, & local_length );
   PLA_Obj_local_width ( destin, & local_width );

   copy_block ( local_length, local_width, typesize,
                brecv, ldims [ rank ],
                bdestin, destin_ldim );

   if ( bsend ) pla_free ( bsend );
   if ( brecv ) pla_free ( brecv );
   pla_free ( ldims );
   pla_free ( counts );
   pla_free ( displs );

   return ( PLA_SUCCESS );
}

int pla_Copy_mat2pmv ( PLA_Obj source, PLA_Obj destin )
{
   int p, owner, rank, size, proj_onto, block_length, offset,
       local_length, local_width, source_ldim, destin_ldim,
       global_length, global_width, global_align_row, global_align_col,
       * ldims = NULL, * counts = NULL, * displs = NULL;
   void * bsource = NULL, * bdestin = NULL,
        * bsend = NULL,   * brecv = NULL;
   MPI_Aint typesize;
   MPI_Datatype datatype = MPI_DATATYPE_NULL;
   MPI_Comm comm = MPI_COMM_NULL;
   PLA_Template template = NULL;
   PLA_Obj view = NULL;

   PLA_Obj_template ( destin, & template );
   PLA_Obj_datatype ( destin, & datatype );
   MPI_Type_extent ( datatype, & typesize );
   PLA_Obj_project_onto ( destin, & proj_onto );

   PLA_Obj_global_length ( source, & global_length );
   PLA_Obj_global_width ( source, & global_width );
   PLA_Obj_global_align_row ( source, & global_align_row );
   PLA_Obj_global_align_col ( source, & global_align_col );

   PLA_Obj_local_buffer ( source, & bsource );
   PLA_Obj_local_buffer ( destin, & bdestin );
   PLA_Obj_local_ldim ( source, & source_ldim );
   PLA_Obj_local_ldim ( destin, & destin_ldim );
   PLA_Obj_local_length ( source, & local_length );
   PLA_Obj_local_width ( source, & local_width );

   if ( proj_onto == PLA_PROJ_ONTO_COL )
   {
      PLA_Temp_comm_row_info ( template, & comm, & rank, & size );
      PLA_Obj_owner_col ( destin, & owner );
   }
   else
   {
      PLA_Temp_comm_col_info ( template, & comm, & rank, & size );
      PLA_Obj_owner_row ( destin, & owner );
   }

   ldims = pla_calloc ( size, sizeof ( int ) );
   counts = pla_calloc ( size, sizeof ( int ) );
   displs = pla_calloc ( size, sizeof ( int ) );

   for ( p = 0 ; p < size ; p++ )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
      {
         ldims [ p ] = local_length;
         PLA_Temp_matrix_distant_width ( template, p,
                                         global_width, global_align_col,
                                         & counts [ p ] );
         counts [ p ] *= local_length;
      }
      else
      {
         PLA_Temp_matrix_distant_length ( template, p,
                                          global_length, global_align_row,
                                          & ldims [ p ] );
         counts [ p ] = ldims [ p ] * local_width;
      }

      if ( p > 0 ) displs [ p ] = displs [ p - 1 ] + counts [ p - 1 ];
   }


   if ( rank == owner )
   {
      PLA_Obj_local_length ( destin, & local_length );
      PLA_Obj_local_width ( destin, & local_width );
      if ( local_length * local_width )
         brecv = pla_calloc ( local_length * local_width, typesize );
   }

   PLA_Obj_local_length ( source, & local_length );
   PLA_Obj_local_width ( source, & local_width );
   if ( local_length * local_width )
      bsend = pla_calloc ( local_length * local_width, typesize );


/* pack */

   copy_block ( local_length, local_width, typesize,
                bsource, source_ldim,
                bsend, ldims [ rank ] );

/* pass message */

/*printf ( "PE%d - gatherv\n", rank );*/

   MPI_Gatherv ( BUFCHK(bsend), counts [ rank ], datatype,
                 BUFCHK(brecv), counts, displs, datatype, owner, comm );

/* unpack */

/*printf ( "PE%d - unpacking\n", rank );*/

   for ( p = 0 ; p < size ; p++ )  counts [ p ] = 0;

   if ( rank == owner )
   {
      if ( proj_onto == PLA_PROJ_ONTO_COL )
         PLA_Temp_matrix_first_col_block ( template, global_width, global_align_col,
                                           & p, & block_length );
      else
         PLA_Temp_matrix_first_row_block ( template, global_length, global_align_row,
                                           & p, & block_length );

      while ( block_length > 0 )
      {

         offset = ( displs [ p ] + counts [ p ] ) * typesize;

         if ( proj_onto == PLA_PROJ_ONTO_COL )
         {
            global_width -= block_length;
            global_align_col += block_length;

            copy_block ( local_length, block_length, typesize,
                         ( char * ) brecv + offset, ldims [ p ],
                         bdestin, destin_ldim );

            bdestin = ( char * ) bdestin + destin_ldim * block_length * typesize;
            counts [ p ] += block_length * ldims [ p ];

            PLA_Temp_matrix_first_col_block ( template, global_width, global_align_col,
                                              & p, & block_length );
        }
        else
        {
           global_length -= block_length;
           global_align_row += block_length;

           copy_block ( block_length, local_width, typesize,
                        ( char * ) brecv + offset, ldims [ p ],
                        bdestin, destin_ldim );

           bdestin = ( char * ) bdestin + block_length * typesize;
           counts [ p ] += block_length;

           PLA_Temp_matrix_first_row_block ( template, global_length, global_align_row,
                                             & p, & block_length );
         }
      }
   }

   if ( bsend ) pla_free ( bsend );
   if ( brecv ) pla_free ( brecv );
   pla_free ( ldims );
   pla_free ( counts );
   pla_free ( displs );

   return ( PLA_SUCCESS );
}


int copy_block ( int length, int width, int typesize,
                 void * source, int ldim_source,
                 void * destin, int ldim_destin )
{
   int i;
   length *= typesize;
   width *= typesize;
   for ( i = 0 ; i < width ; i += typesize )
      memcpy ( ( char * ) destin + i * ldim_destin,
               ( char * ) source + i * ldim_source, length );

   return ( 0 );
}
