/**************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/


#include "mpi.h"
#include "PLA.h"
#include <math.h>
#ifndef FLOAT_EPSILON
#define FLOAT_EPSILON 1.0e-3
#endif
#ifndef DOUBLE_EPSILON
#define DOUBLE_EPSILON 1.0e-6
#endif

/*****************************************************************
Scalable BLAS test suite utility routine: PLA_test_dextract


Matrix AA, stored in array a_all, holds the global matrix.
Upon return, array A, stored in array *a, holds the
local part of the matrix, where it is assumed that AA
was alligned as follows:
Assume template matrix and r x c processor grid
			 / X00  X01 ... \
	  X =  | X10  X11 ... |    with Xij  nb x nb assigned to
			 | X20  X21 ... |       processor (i%r,j%c)
			 \ ...  ... ... /
parameters ia, ja indicate how matrix A is aligned to this
template, and hence dictates how the local block is to
be extracted.

Parameters

lda_all           INPUT: int
						leading dimensions of global arrays AA

*llda             OUTPUT: pointer to int
						pointer to leading dimension of array that
						holds A

Copyright 1996  The University of Texas at Austin
This code is distributed under gnu license.

Programmed by Robert van de Geijn
				  rvdg@cs.utexas.edu
********************************************************************/

int PLA_test_compare ( void * a_all, int lda_all, PLA_Obj a_obj )
{
   PLA_Obj_obj_type Object_Type;
   MPI_Datatype      Object_Datatype;
   int               g_length, g_width, g_align_row, g_align_col,
                     l_length, l_width, l_stride, l_ldim;
   PLA_Template   Object_Template;
   void        * Object_Buffer;
   int	         my_row, my_col, ownerrow, ownercol, me1D, owner1d;
   double        val1, val2, difference;
   double        val3 = 0.0, val4 = 0.0, difference2 = 0.0;
   int           i, j, local_i, local_j;
   int           mismatches = 0;
   int           project, owner_row, owner_col;
   char * message = "Compare ";

   PLA_Obj_objtype(a_obj, &Object_Type); 
   PLA_Obj_datatype(a_obj, &Object_Datatype);
   PLA_Obj_template(a_obj, &Object_Template);
   PLA_Obj_global_info ( a_obj,    & g_length,
                                   & g_width,
                                   & project,
                                   & owner_row,
                                   & owner_col,
                                   & g_align_row,
                                   & g_align_col);

   PLA_Obj_objtype    (a_obj, &Object_Type) ;
   PLA_Obj_datatype   (a_obj, &Object_Datatype);
   PLA_Obj_local_info ( a_obj, & l_length,
                                  & l_width,
                                  & Object_Buffer,
                                  & l_stride,
                                  & l_ldim );

   if ( Object_Datatype != MPI_DOUBLE &&
        Object_Datatype != MPI_FLOAT &&
        Object_Datatype != MPI_COMPLEX &&
        Object_Datatype != MPI_DOUBLE_COMPLEX &&
        Object_Datatype != MPI_INT )
   {
      printf ( "Object Datatype not valid in PLA_Test_compare!!!\n" );
      return ( -1 );
   }


   switch ( Object_Type )
   {
      case PLA_MATRIX :

         PLA_Temp_mycoords ( Object_Template, & my_row, & my_col );

         for ( i = 0 ; i < g_length ; i++ )
         {
            for ( j = 0 ; j < g_width ; j++ )
            {

               PLA_Temp_matrix_owner_rowcol ( Object_Template, 
                                             i + g_align_row, j + g_align_col,
                                             & ownerrow, & ownercol );

               if ( my_row == ownerrow && my_col == ownercol )
               {
                  PLA_Temp_matrix_local_length_width ( Object_Template,
                                                      i, j,
                                                      g_align_row, g_align_col,
                                                      & local_i, & local_j );


                  if ( Object_Datatype == MPI_DOUBLE )
                  {
                     val1 = * ( ( double * ) Object_Buffer + local_i + local_j * l_ldim );
                     val2 = * ( ( double * ) a_all + i + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_DOUBLE_COMPLEX )
                  {
              val1 =   *((double*)Object_Buffer+ local_i*2 +local_j*l_ldim*2 ); 
              val2 =   * ( ( double * ) a_all + i*2 + j * lda_all );
              val3 =   *((double*)Object_Buffer+local_i*2+1+local_j*l_ldim*2 ); 
              val4 =   *((double*) a_all + i*2+1 + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_FLOAT )
                  {
                     val1 = * ( ( float * ) Object_Buffer + local_i + local_j * l_ldim );
                     val2 = * ( ( float * ) a_all + i + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_INT )
                  {
                     val1 = * ( ( int * ) Object_Buffer + local_i + local_j * l_ldim );
                     val2 = * ( ( int * ) a_all + i + j * lda_all );
                  }

                  difference = fabs ( (val2 - val1)/val2 );

                  if( ( Object_Datatype == MPI_DOUBLE && difference > DOUBLE_EPSILON ) ||
                      ( Object_Datatype != MPI_DOUBLE && difference > FLOAT_EPSILON  ) )
                  {
                     mismatches++;
                     printf ( "%s : PE(%d,%d) glob(%d,%d,%d,%d) loc(%d,%d) %f vs %f\n",
                              message, my_row, my_col, i, j, g_align_row, g_align_col,
                              local_i, local_j, val1, val2 );
                  }
               }
            }
         }

         break;

      case PLA_MVECTOR :

         PLA_Temp_me1D ( Object_Template, & me1D );

         for ( i = 0 ; i < g_length ; i++ )
         {
            PLA_Temp_vector_owner_rank ( Object_Template, i + g_align_row, & owner1d );

            if ( me1D == owner1d )
            {
               PLA_Temp_vector_local_length ( Object_Template,
                                             i, g_align_row, & local_i );

               for ( j = 0 ; j < g_width ; j++ )
               {
                  local_j = j;

                  if ( Object_Datatype == MPI_DOUBLE )
                  {
                     val1 = * ( ( double * ) Object_Buffer + local_i + j * l_ldim );
                     val2 = * ( ( double * ) a_all + i + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_FLOAT )
                  {
                     val1 = * ( ( float * ) Object_Buffer + local_i + j * l_ldim );
                     val2 = * ( ( float * ) a_all + i + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_INT )
                  {
                     val1 = * ( ( int * ) Object_Buffer + local_i + j * l_ldim );
                     val2 = * ( ( int * ) a_all + i + j * lda_all );
                  }

                  difference = fabs ( (val2 - val1)/val2 );
                  difference2 = fabs ((val4 - val3)/val4 );

                  if( ( ((Object_Datatype == MPI_DOUBLE) || 
                         (Object_Datatype == MPI_DOUBLE_COMPLEX)) &&
                       ((difference  > DOUBLE_EPSILON ) ||
                        (difference2 > DOUBLE_EPSILON )) )            ||
                     ( ((Object_Datatype == MPI_FLOAT) || 
                         (Object_Datatype == MPI_COMPLEX)) &&
                       ((difference  > FLOAT_EPSILON ) ||
                        (difference2 > FLOAT_EPSILON )) )              )

                  {
                     mismatches++;
                     printf ( "%s : PE(%d) glob(%d,%d,%d,%d) loc(%d,%d) %f vs %f\n",
                              message, me1D, i, j, g_align_row, g_align_col,
                              local_i, local_j, val1, val2 );
                  }
               }
            }
         }

         break;

      case PLA_VECTOR :

         local_j = j = 1;
         PLA_Temp_me1D ( Object_Template, & me1D );

         for ( i = 0 ; i < g_length ; i++ )
         {
            PLA_Temp_vector_owner_rank ( Object_Template, i + g_align_row, & owner1d );

            if ( me1D == owner1d )
            {
               PLA_Temp_vector_local_length ( Object_Template,
                                             i, g_align_row, & local_i );

               if ( Object_Datatype == MPI_DOUBLE )
               {
                  val1 = * ( ( double * ) Object_Buffer + local_i );
                  val2 = * ( ( double * ) a_all + i );
               }
               else if ( Object_Datatype == MPI_FLOAT )
               {
                  val1 = * ( ( float * ) Object_Buffer + local_i );
                  val2 = * ( ( float * ) a_all + i );
               }
               else if ( Object_Datatype == MPI_INT )
               {
                  val1 = * ( ( int * ) Object_Buffer + local_i );
                  val2 = * ( ( int * ) a_all + i );
               }

               difference = fabs ( (val2 - val1)/val2 );

                  if( ( Object_Datatype == MPI_DOUBLE && difference > DOUBLE_EPSILON ) ||
                      ( Object_Datatype != MPI_DOUBLE && difference > FLOAT_EPSILON  ) )
               {
                  mismatches++;
                  printf ( "%s : PE(%d) glob(%d,%d,%d,%d) loc(%d,%d) %f vs %f (%e)\n",
                           message, me1D, i, j, g_align_row, g_align_col,
                           local_i, local_j, val1, val2, difference );
               }
            }
         }

         break;

      default :

         printf("Failure in compare, invalid object type\n");
         return BAD_OBJ_TYPE;
   }

   return mismatches;
}

int PLA_Test_gdist_to_local ( void * a_all, int lda_all, PLA_Obj a_obj )
{
   PLA_Obj_obj_type Object_Type;
   MPI_Datatype      Object_Datatype;
   int               g_length, g_width, g_align_row, g_align_col,
                     l_length, l_width, l_stride, l_ldim;
   PLA_Template       Object_Template;
   void            * Object_Buffer;
   int	             my_row, my_col, ownerrow, ownercol, me1D, owner1d;
   double            val1, val2, difference;
   int               i, j, local_i, local_j;
   int           project, owner_row, owner_col;
int me;

MPI_Comm_rank(MPI_COMM_WORLD, &me);

   PLA_Obj_objtype(a_obj, &Object_Type); 
   PLA_Obj_datatype(a_obj, &Object_Datatype);
   PLA_Obj_template(a_obj, &Object_Template);
   PLA_Obj_global_info ( a_obj,    & g_length,
                                   & g_width,
                                   & project,
                                   & owner_row,
                                   & owner_col,
                                   & g_align_row,
                                   & g_align_col);

   PLA_Obj_objtype (a_obj, &Object_Type) ;
   PLA_Obj_datatype  (a_obj, &Object_Datatype);
   PLA_Obj_local_info ( a_obj, & l_length,
                                  & l_width,
                                  & Object_Buffer,
                                  & l_stride,
                                  & l_ldim );

   switch ( Object_Type )
   {
      case PLA_MATRIX :

         PLA_Temp_mycoords ( Object_Template, & my_row, & my_col );

         for ( i = 0 ; i < g_length ; i++ )
            for ( j = 0 ; j < g_width ; j++ )
            {

               PLA_Temp_matrix_owner_rowcol ( Object_Template, 
                                             i + g_align_row, j + g_align_col,
                                             & ownerrow, & ownercol );

               if ( my_row == ownerrow && my_col == ownercol )
               {
                  PLA_Temp_matrix_local_length_width ( Object_Template,
                                                      i, j,
                                                      g_align_row, g_align_col,
                                                      & local_i, & local_j );
/*************************************************************************/
                  if ( Object_Datatype == MPI_DOUBLE )
                     *((double*) Object_Buffer + local_i + local_j * l_ldim ) =
                     * ( ( double * ) a_all + i + j * lda_all );
                  else if ( Object_Datatype == MPI_FLOAT )
                     *(( float * )Object_Buffer + local_i + local_j * l_ldim ) =
                     *(( float * )a_all + i + j * lda_all );
/* The weird Mult by 2 is because even though the OBJECTS "know" that   */
/* they are complex and those parameters fetched from them reflect that */
/* fact, I am treating them like pairs of floats or doubles.            */
                  else if ( Object_Datatype == MPI_DOUBLE_COMPLEX )
                  {
                     *((double*)Object_Buffer+ local_i*2   +local_j*l_ldim*2 ) =
                     * ( ( double * ) a_all + i*2 + j * lda_all );
                     *((double*)Object_Buffer+ local_i*2+1 +local_j*l_ldim*2 ) =
                     * ( ( double * ) a_all + i*2+1 + j * lda_all );
                  }
                  else if ( Object_Datatype == MPI_INT )
                     *( ( int * ) Object_Buffer + local_i + local_j * l_ldim ) =
                     *( ( int * ) a_all + i + j * lda_all );
/*************************************************************************/
               }
            }

         break;

      case PLA_MVECTOR :

         PLA_Temp_me1D ( Object_Template, & me1D );

         for ( i = 0 ; i < g_length ; i++ )
         {
            PLA_Temp_vector_owner_rank ( Object_Template, i + g_align_row, & owner1d );

            if ( me1D == owner1d )
            {
               PLA_Temp_vector_local_length ( Object_Template,
                                             i, g_align_row, & local_i );

               for ( j = 0 ; j < g_width ; j++ )
               {
                  if ( Object_Datatype == MPI_DOUBLE )
                     * ( ( double * ) Object_Buffer + local_i + j * l_ldim ) = 
                     * ( ( double * ) a_all + i + j * lda_all );
                  else if ( Object_Datatype == MPI_FLOAT )
                     * ( ( float * ) Object_Buffer + local_i + j * l_ldim ) =
                     * ( ( float * ) a_all + i + j * lda_all );
                  else if ( Object_Datatype == MPI_INT )
                     * ( ( int * ) Object_Buffer + local_i + j * l_ldim ) = 
                     * ( ( int * ) a_all + i + j * lda_all );
               }
            }
         }

         break;

      case PLA_VECTOR :
      
         PLA_Temp_me1D ( Object_Template, & me1D );

         for ( i = 0 ; i < g_length ; i++ )
         {
            PLA_Temp_vector_owner_rank ( Object_Template, i + g_align_row, & owner1d );

            if ( me1D == owner1d )
            {
               PLA_Temp_vector_local_length ( Object_Template,
                                             i, g_align_row, & local_i );

               if ( Object_Datatype == MPI_DOUBLE )
                  * ( ( double * ) Object_Buffer + local_i ) = 
                     * ( ( double * ) a_all + i );
               else if ( Object_Datatype == MPI_FLOAT )
                  * ( ( float *) Object_Buffer + local_i ) = 
                     * ( ( float * ) a_all + i );
               else if ( Object_Datatype == MPI_INT )
                  * ( ( int * ) Object_Buffer + local_i ) =
                     * ( ( int * ) a_all + i );
            }
         }

         break;

      default:
         printf("Failure in dextract, invalid object type\n");
         printf("Object type found: %d\n", Object_Type);
         return BAD_OBJ_TYPE; /* Well, we need a name */
   }

   return ( 0 );
}

