/***************************************************************************
  03 April 1996 10:11
  
  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu

  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu
  
  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu
  
  Greg Morrow, Department of Physics, 
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu
 
  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu
 
***************************************************************************/





#include "mpi.h"
#include "PLA.h"

extern int PLA_ALGORYTHMIC_BLOCKSIZE;


void PLA_test_dsymm( PLA_Template template, char *side, char *uplo,
			 int m, int n, 
			 double alpha, double beta,
			 int ia, int ja, int ib, int jb, int ic, int jc,
			 double *time, double *mflops, int *info )

/*****************************************************************
PLA_library DSYMM test routine

Programmed by Robert van de Geijn and John Gunnels
				  rvdg@cs.utexas.edu      gunnels@cs.utexas.edu

Modified from sB_test_PLA_dgemm.c, which is distributed under gnu
licence
********************************************************************/
{
  int mym, myn,         /* this node's row and column dimension    */
	 lda_all, ldb_all,   /* leading dimensions of global matrices   */
	 ldc_all,
	 llda, lldb, lldc,   /* leading dimensions of local matrices    */
	 idummy,             /* dummy variable                          */
	 ii;                 /* loop counter -- dummy                   */
  double
	 *a_all, *b_all,     /* arrays holding the global matrices      */
	 *c_all,
	 *a, *b, *c;         /* arrays holding the local matrices       */

  PLA_Obj a_obj=NULL, b_obj=NULL, c_obj=NULL;
  PLA_Obj Alpha_Mscalar = NULL, Beta_Mscalar = NULL;
  MPI_Comm comm1d;
  MPI_Datatype   datatype;
  int      np, ALG_NB_SIZE;
  int      side_int, uplo_int;
 
  if ('R' == *side)
     side_int = PLA_SIDE_RIGHT;
  else side_int = PLA_SIDE_LEFT;

  if ('U' == *uplo)
     uplo_int = PLA_UPPER_TRIANGULAR;
  else uplo_int = PLA_LOWER_TRIANGULAR;

	/* Create distributed matrix objects for A, B, C */
  c_all = (double *) pla_malloc( m * n * sizeof( double ) );
  ldc_all = m;
  PLA_Matrix_create ( MPI_DOUBLE, m, n, template, ic, jc, &c_obj );
  /* Fill C in with all zeros */
  PLA_test_dgenmatrix( "S", 1.0, m, n, c_all, ldc_all );
  PLA_Test_gdist_to_local( c_all, ldc_all, c_obj );

  if ( *side == 'L' ) {
         a_all = (double *) pla_malloc( m * m * sizeof( double ) );
	 PLA_Matrix_create ( MPI_DOUBLE, m, m, template, ia, ja, &a_obj );
	 lda_all = m;
	 PLA_test_dgenmatrix( "S", 1000.0, m, m, a_all, lda_all );
	 PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
  }
  else {
         a_all = (double *) pla_malloc( n * n * sizeof( double ) );
	 PLA_Matrix_create ( MPI_DOUBLE, n, n, template, ia, ja, &a_obj );
	 lda_all = n;
	 PLA_test_dgenmatrix( "S", 100.0, n, n, a_all, lda_all );
	 PLA_Test_gdist_to_local( a_all, lda_all, a_obj );
  }


  b_all = (double *) pla_malloc( m * n * sizeof( double ) );
  PLA_Matrix_create ( MPI_DOUBLE, m, n, template, ib, jb, &b_obj );
  ldb_all = m;
  PLA_test_dgenmatrix( "S", 10.00000, m, n, b_all, ldb_all );
  PLA_Test_gdist_to_local( b_all, ldb_all, b_obj );
		/* compute correct sequential answer             */
   PLA_dsymm( side, uplo, &m, &n, &alpha, a_all, &lda_all,
              b_all, &ldb_all, &beta,  c_all, &ldc_all );
	/* compute and time parallel answer              */

  ALG_NB_SIZE = PLA_ALGORYTHMIC_BLOCKSIZE;

  PLA_Obj_datatype(a_obj, &datatype);
  PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS,
                           1, 1, template, &Alpha_Mscalar);
  PLA_Obj_set_contents(Alpha_Mscalar, &alpha);
  PLA_Mscalar_create(datatype, PLA_ALL_ROWS, PLA_ALL_COLS,
                           1, 1, template, &Beta_Mscalar);
  PLA_Obj_set_contents(Beta_Mscalar, &beta);

  MPI_Barrier( MPI_COMM_WORLD );
  *time = MPI_Wtime();
  
  PLA_Symm(side_int,uplo_int,Alpha_Mscalar,a_obj, b_obj, Beta_Mscalar,c_obj);

  MPI_Barrier( MPI_COMM_WORLD );

  PLA_Temp_comm1D( template, &comm1d );
  MPI_Comm_size( comm1d, &np );

  *time = MPI_Wtime() - *time;
						/* compute MFLOPS/node                           */
  *mflops = 1.0 * m * m *n   / *time * 1.0e-6 / np;
						/* check answer                                  */

  *info = PLA_test_compare( c_all, ldc_all, c_obj);
						/* free space                                    */
  if ( a_all != NULL ) pla_free( a_all );
  if ( b_all != NULL ) pla_free( b_all );
  if ( c_all != NULL ) pla_free( c_all );
  PLA_Obj_free( &a_obj );
  PLA_Obj_free( &b_obj);
  PLA_Obj_free( &c_obj );
  PLA_Obj_free ( & Alpha_Mscalar );
  PLA_Obj_free ( & Beta_Mscalar );


}




