/***************************************************************************
  03 April 1996 10:11

  Copyright (c) 1996 Robert van de Geijn and The University of Texas at Austin.
  See the file GNU_LICENSE for details on the gnu license

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 1, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Written by:   Many People including:

  Gregory A. Baker, Center for Space Research,
  University of Texas at Austin  78722      baker@csr.utexas.edu

  Philip A. Alpatov, Department of Physics,
  University of Texas at Austin 78712.     philip@physics.utexas.edu


  James Overfelt, Department of Mathematics,
  University of Texas at Austin  78712.    overfelt@@math.utexas.edu

  John Andrew Gunnels, Department of Computer Sciences,
  University of Texas at Austin 78712.     gunnels@@cs.utexas.edu

  Greg Morrow, Department of Physics,
  University of Texas at Austin 78712.     morrow@@physics.utexas.edu

  Dr. Robert Van de Geijn, Department of Computer Sciences,
  University of Texas at Austin 78712.     rvdg@@cs.utexas.edu

******************************************************************************/
#include "PLA.h"

/* This is really just used to get around the testing stuff */
/* treat a row as a column and a column as a row temporarily*/
/* the re-transpose it (via the same call) later.  NO data  */
/* is moved, obviously the fields are just swapped.         */

int PLA_Obj_implicit_local_transpose(PLA_Obj object)
{
   int temp;

   temp = object->local_stride;
   object->local_stride = object-> local_lld;
   object->local_lld = temp;

   temp = object->local_width;
   object->local_width = object->local_length;
   object->local_length = temp;
   return PLA_SUCCESS;
}

PLA_Obj_clone(PLA_Obj Y, PLA_Obj *Conformal_to_Y)
{
   int object_type_Y, owner_row_Y, owner_col_Y, Projection_Y;
   int width_Y, length_Y;

   PLA_Obj_objtype( Y, &object_type_Y);
   if(PLA_MATRIX == object_type_Y)
      PLA_Matrix_create_conf_to( Y, Conformal_to_Y);
   else if (PLA_MSCALAR == object_type_Y)
   {
      PLA_Obj_owner_row(Y, &owner_row_Y);
      PLA_Obj_owner_col(Y, &owner_col_Y);
      PLA_Mscalar_create_conf_to( Y, owner_row_Y, owner_col_Y, Conformal_to_Y);
   }
   else if((PLA_PVECTOR == object_type_Y)|| (PLA_PMVECTOR == object_type_Y))
   {
      PLA_Obj_project_onto(Y, &Projection_Y);
      if(PLA_PROJ_ONTO_ROW == Projection_Y)
      {
         PLA_Obj_owner_row(Y, &owner_row_Y);
         PLA_Obj_global_length(Y, &length_Y);
         PLA_Pmvector_create_conf_to(Y, Projection_Y,   owner_row_Y, length_Y,
                                     Conformal_to_Y);
      }
      else if(PLA_PROJ_ONTO_COL == Projection_Y)
      {
         PLA_Obj_owner_col(Y, &owner_col_Y);
         PLA_Obj_global_width(Y, &width_Y);
         PLA_Pmvector_create_conf_to(Y, Projection_Y,   owner_col_Y, width_Y,
                                     Conformal_to_Y);
      }
   }
   else /* The enter routine assured us a legitimate object type */
   {
      PLA_Obj_global_width(Y, &width_Y);
      PLA_Mvector_create_conf_to(Y, width_Y, Conformal_to_Y);
   }
   return(PLA_SUCCESS);
}

/* Overwrites the contents with the sqrt of the contents -- currently */
/* only for float and double */
int   PLA_Take_sqrt(PLA_Obj obj )
{
   int loc_len, loc_wid;
   MPI_Datatype datatype;
   void         *data;

   PLA_Obj_local_length ( obj, & loc_len );
   PLA_Obj_local_width ( obj, & loc_wid );
   PLA_Obj_datatype(obj, &datatype);
   PLA_Obj_local_buffer(obj, &data);
   if ( loc_len && loc_wid )
      if (MPI_DOUBLE == datatype) 
         *(double*)data = sqrt(*(double*)data);
      else if (MPI_FLOAT == datatype) 
         *(float*)data = sqrt(*(float*)data);
      else printf("Attempt to take sqrt of complex number - not yet supported\n");
   return(PLA_SUCCESS);
}

int PLA_Obj_alloc_type(PLA_Obj obj, int *allocation_type)
{
   *allocation_type = obj->allocation_type;
   return(PLA_SUCCESS);
}

int PLA_Obj_data_space_state(PLA_Obj obj, int *space_state)
{
   *space_state = obj->has_data_space;
   return(PLA_SUCCESS);
}

