/* test program */

#include <stdio.h>
#include "mpi.h"
#include "PLA.h"
#include "PRISM.h"

#define PRISM_PRT_IN

void prism_pla_mat_gen(PLA_Obj A);

int nb,tnb,cut,r,s,n;

main(int argc, char **argv )
 
{
  int  ndims, dims[2];  /* for 2-D MPI_Comm creation               */
  int  periods[2];      /* S.A.A. Do we wrap the Comm. mesh        */
  int  reorder;         /* Allow/disallow re-ordering of proc #s   */
  MPI_Comm comm2d;      /* The 2-D communicator used               */
  PLA_Template template=NULL;
  int j;
  double prism_time, ave_time, max_time, temp_time;

  FILE *infile;

  PLA_Obj A=NULL,Z=NULL,D=NULL;

  /* Initialize MPI */
  MPI_Init( &argc, &argv );
  MPI_Comm_rank(MPI_COMM_WORLD, &me );
  
  /* Extract data from input file */ 
  if (me == 0) {
    infile = fopen(argv[1], "r");
    if (infile == (FILE *)NULL) 
      printf("driver: input file = %s could not be opened", argv[1]);
    fscanf(infile, "%i %i %i %i %i %i", &r, &s, &n, &tnb, &nb, &cut);
    fclose(infile);
    printf("\nget mesh=%dx%d n=%d tnb=%d nb=%d cut=%d\n",r,s,n,tnb,nb,cut);
  }

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Bcast( &r, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &s, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &n, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &nb, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &tnb, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &cut, 1, MPI_INT, 0, MPI_COMM_WORLD );

  /* Create 2D communicator -- make a function */
  ndims = 2;  /* 2-Dimensional Communicator */
  dims[0] = r; dims[1] = s;  /* r x c grid */
  periods[0] = TRUE; periods[1] = TRUE;  /* Torus */
  reorder = FALSE;  /* dis-allows processors to be re-ordered */
  MPI_Cart_create(MPI_COMM_WORLD, ndims, dims, periods, reorder, &comm2d);

  /* Initialize timing */
  for (j=0; j<PRISM_TIME; j++) t_syisda[j]=0.0;

  /* Initialize template */
  PLA_Init (comm2d);
  PLA_Temp_create(tnb, 0, &template );
  MPI_Comm_rank(comm2d, &me );

  /* Initialize TEST matrix A */
  PLA_Matrix_create(MPI_DOUBLE, n, n, template, 0, 0, &A);
  PLA_Obj_set_to_zero(A);
  prism_pla_mat_gen(A); 

  /* Create global vector matrix Z */
  PLA_Matrix_create_conf_to( A, &Z );

  /* Create global eigenvalue multivector D */
  PLA_Mvector_create(MPI_DOUBLE, n, 1, template, 0, &D);

  /* Run SYISDA algorithm */
  PRISM_syisda(A, Z, D, cut);

  /* Print results */
  if (me==ROOT) printf("start PRISM_print_result\n\n");
  PRISM_print_result(A, Z, D);

  prism_time=t_syisda[PRISM_TIME-1];

  MPI_Reduce(&prism_time, &max_time, 1, MPI_DOUBLE, MPI_MAX, 0, comm2d);
  MPI_Reduce(&prism_time, &temp_time, 1, MPI_DOUBLE, MPI_SUM, 0, comm2d);
  ave_time = temp_time / (r*s);
  
  for (j=0; j<PRISM_TIME-1; j++) {
    MPI_Reduce(t_syisda+j, &temp_time, 1, MPI_DOUBLE, MPI_MAX, 0, comm2d);
    if (me==0) t_syisda[j]=temp_time;
  }

  temp_time=0.0;
  for (j=0; j<PRISM_TIME-1; j++) temp_time+=t_syisda[j];
    
  if (me==0) {
    printf("\nsetup=%8.4f range=%8.4f shift=%8.4f beta=%8.4f\n",
	   t_syisda[t_setup], t_syisda[t_range], t_syisda[t_shift], 
	   t_syisda[t_beta]);
    printf("sbr  =%8.4f split=%8.4f end  =%8.4f max_sum=%11.4e\n",
	   t_syisda[t_sbr], t_syisda[t_split], t_syisda[t_end], temp_time);
    printf("local time: %11.4e   max time: %11.4e  ave time: %11.4e \n\n",
             prism_time, max_time, ave_time);
  }

  /* free PLAPACK objects */
  PLA_Obj_free(&A);
  PLA_Obj_free(&Z);
  PLA_Obj_free(&D);

  /* clean up MPI */
  MPI_Comm_free(&comm2d);

  /* PLA_Finalize(); */
  MPI_Finalize();

}
 

void prism_pla_mat_gen(PLA_Obj A)
{
  void  *buf_localA=NULL;
  MPI_Datatype datatype;
  int i,j,k,l, g_i,g_j, ma,na,i_ldA;

  PLA_Matrix_local_all( A, &datatype,&ma,&na,&buf_localA, &i_ldA);
  g_j=(me%s)*(r*tnb)-1;
  for (j=0; j<na; j++) {
    if (j%(r*tnb)==0 && j!=0)
      g_j=g_j+(s-1)*r*tnb+1;
    else
      g_j++;
    l=j*i_ldA;    
    g_i=((int)(me/s))*(tnb)-1;
    for(i=0; i<ma; i++) {
      if(i%tnb == 0 && i!=0)
        g_i+=(r*tnb-tnb+1);
      else
        g_i++;

      /* Diagonal matrix */
      /*
      if (g_i == g_j)
        *(((double *)buf_localA)+l+i)= (double)((g_i+1)%2); 
      else
        *(((double *)buf_localA)+l+i)= 0.0;
      */
      
      /* indexing 
       *(((double *)buf_localA)+l+i)=(double)(g_i+1)+ 0.1*(double)(g_j+1);
       */


      /* Uniformly dist. I */
      if (g_i == g_j)
        *(((double *)buf_localA)+l+i)= (double)(g_i+1)+ 0.1*(double)(g_i+1); 
      else
        *(((double *)buf_localA)+l+i)= (double)(g_i+g_j+2)/((g_i+1)*(g_j+1));

      /* Uniformly dist. II */
      if (g_i == g_j)
        *(((double *)buf_localA)+l+i)= (double)(g_i+1); 
      else
        *(((double *)buf_localA)+l+i)= (double)(g_i+g_j+2)/((g_i+1)*(g_j+1));
    }
  }
}


void prism_pla_matI_gen(PLA_Obj A)
{
  void  *buf_localA=NULL;
  MPI_Datatype datatype;
  int i,j,k,l, g_i,g_j, ma,na,i_ldA;

  PLA_Matrix_local_all( A, &datatype,&ma,&na,&buf_localA, &i_ldA);
  PLA_Zero_loc_obj(A);

  g_j=(me%s)*(r*tnb)-1;
  for (j=0; j<na; j++) {
    if (j%(r*tnb)==0 && j!=0)
      g_j=g_j+(s-1)*r*tnb+1;
    else
      g_j++;
    l=j*i_ldA;    
    g_i=((int)(me/s))*(tnb)-1;
    for(i=0; i<ma; i++) {
      if(i%tnb == 0 && i!=0)
        g_i+=(r*tnb-tnb+1);
      else
        g_i++;

      if (g_i==g_j)
	*(((double *)buf_localA)+l+i)=(double)1.0;
    }
  }
}



